@@ -315,13 +315,13 @@ void CL_ParseEntityLump(char *entdata)
 	data = entdata;
 	if (!data)
 		return;
-	if (!COM_ParseToken(&data, false))
+	if (!COM_ParseTokenConsole(&data))
 		return; // error
 	if (com_token[0] != '{')
 		return; // error
 	while (1)
 	{
-		if (!COM_ParseToken(&data, false))
+		if (!COM_ParseTokenConsole(&data))
 			return; // error
 		if (com_token[0] == '}')
 			break; // end of worldspawn
@@ -331,7 +331,7 @@ void CL_ParseEntityLump(char *entdata)
 			strlcpy (key, com_token, sizeof (key));
 		while (key[strlen(key)-1] == ' ') // remove trailing spaces
 			key[strlen(key)-1] = 0;
-		if (!COM_ParseToken(&data, false))
+		if (!COM_ParseTokenConsole(&data))
 			return; // error
 		strlcpy (value, com_token, sizeof (value));
 		if (!strcmp("sky", key))
@@ -492,7 +492,7 @@ static void QW_CL_RequestNextDownload(void)
 		// now that we have a world model, set up the world entity, renderer
 		// modules and csqc
 		cl.entities[0].render.model = cl.worldmodel = cl.model_precache[1];
-		CL_BoundingBoxForEntity(&cl.entities[0].render);
+		CL_UpdateRenderEntity(&cl.entities[0].render);
 
 		R_Modules_NewMap();
 
@@ -1017,7 +1017,8 @@ void CL_ParseServerInfo (void)
 
 		// parse player number
 		i = MSG_ReadByte();
-		cl.qw_spectator = (i & 128) != 0;
+		// cl.qw_spectator is an unneeded flag, cl.scores[cl.playerentity].qw_spectator works better (it can be updated by the server during the game)
+		//cl.qw_spectator = (i & 128) != 0;
 		cl.playerentity = cl.viewentity = (i & 127) + 1;
 		cl.scores = (scoreboard_t *)Mem_Alloc(cls.levelmempool, cl.maxclients*sizeof(*cl.scores));
 
@@ -1157,7 +1158,7 @@ void CL_ParseServerInfo (void)
 
 		// we now have the worldmodel so we can set up the game world
 		cl.entities[0].render.model = cl.worldmodel = cl.model_precache[1];
-		CL_BoundingBoxForEntity(&cl.entities[0].render);
+		CL_UpdateRenderEntity(&cl.entities[0].render);
 		R_Modules_NewMap();
 	}
 
@@ -1369,7 +1370,6 @@ void CL_ParseClientdata (void)
 		cl.stats[STAT_ITEMS] = MSG_ReadLong ();
 
 	cl.onground = (bits & SU_ONGROUND) != 0;
-	csqc_onground = cl.onground;	//[515]: cause without this csqc will receive not right value on svc_print =/
 	cl.inwater = (bits & SU_INWATER) != 0;
 
 	if (cls.protocol == PROTOCOL_DARKPLACES5)
@@ -1412,6 +1412,9 @@ void CL_ParseClientdata (void)
 
 	// viewzoom interpolation
 	cl.mviewzoom[0] = (float) max(cl.stats[STAT_VIEWZOOM], 2) * (1.0f / 255.0f);
+
+	// force a recalculation of the player prediction
+	cl.movement_replay = true;
 }
 
 /*
@@ -1438,14 +1441,12 @@ void CL_ParseStatic (int large)
 	ent->render.skinnum = ent->state_baseline.skin;
 	ent->render.effects = ent->state_baseline.effects;
 	ent->render.alpha = 1;
-	//ent->render.scale = 1;
 
 	//VectorCopy (ent->state_baseline.origin, ent->render.origin);
 	//VectorCopy (ent->state_baseline.angles, ent->render.angles);
 
 	Matrix4x4_CreateFromQuakeEntity(&ent->render.matrix, ent->state_baseline.origin[0], ent->state_baseline.origin[1], ent->state_baseline.origin[2], ent->state_baseline.angles[0], ent->state_baseline.angles[1], ent->state_baseline.angles[2], 1);
-	Matrix4x4_Invert_Simple(&ent->render.inversematrix, &ent->render.matrix);
-	CL_BoundingBoxForEntity(&ent->render);
+	CL_UpdateRenderEntity(&ent->render);
 
 	// This is definitely cheating...
 	if (ent->render.model == NULL)
@@ -2043,6 +2044,133 @@ void CL_ParseTempEntity(void)
 	}
 }
 
+// look for anything interesting like player IP addresses or ping reports
+qboolean CL_ExaminePrintString(const char *text)
+{
+	const char *t;
+	if (!strcmp(text, "Client ping times:\n"))
+	{
+		cl.parsingtextmode = CL_PARSETEXTMODE_PING;
+		for(cl.parsingtextplayerindex = 0; cl.parsingtextplayerindex < cl.maxclients && !cl.scores[cl.parsingtextplayerindex].name[0]; cl.parsingtextplayerindex++)
+			;
+		if (cl.parsingtextplayerindex >= cl.maxclients) // should never happen, since the client itself should be in cl.scores
+		{
+			Con_Printf("ping reply but empty scoreboard?!?\n");
+			cl.parsingtextmode = CL_PARSETEXTMODE_NONE;
+			cl.parsingtextexpectingpingforscores = 0;
+		}
+		cl.parsingtextexpectingpingforscores = cl.parsingtextexpectingpingforscores ? 2 : 0;
+		return !cl.parsingtextexpectingpingforscores;
+	}
+	if (!strncmp(text, "host:    ", 9))
+	{
+		// cl.parsingtextexpectingpingforscores = false; // really?
+		cl.parsingtextmode = CL_PARSETEXTMODE_STATUS;
+		cl.parsingtextplayerindex = 0;
+		return true;
+	}
+	if (cl.parsingtextmode == CL_PARSETEXTMODE_PING)
+	{
+		// if anything goes wrong, we'll assume this is not a ping report
+		qboolean expected = cl.parsingtextexpectingpingforscores;
+		cl.parsingtextexpectingpingforscores = 0;
+		cl.parsingtextmode = CL_PARSETEXTMODE_NONE;
+		t = text;
+		while (*t == ' ')
+			t++;
+		if ((*t >= '0' && *t <= '9') || *t == '-')
+		{
+			int ping = atoi(t);
+			while ((*t >= '0' && *t <= '9') || *t == '-')
+				t++;
+			if (*t == ' ')
+			{
+				int charindex = 0;
+				t++;
+				if(cl.parsingtextplayerindex < cl.maxclients)
+				{
+					for (charindex = 0;cl.scores[cl.parsingtextplayerindex].name[charindex] == t[charindex];charindex++)
+						;
+					if (cl.scores[cl.parsingtextplayerindex].name[charindex] == 0 && t[charindex] == '\n')
+					{
+						cl.scores[cl.parsingtextplayerindex].qw_ping = bound(0, ping, 9999);
+						for (cl.parsingtextplayerindex++;cl.parsingtextplayerindex < cl.maxclients && !cl.scores[cl.parsingtextplayerindex].name[0];cl.parsingtextplayerindex++)
+							;
+						//if (cl.parsingtextplayerindex < cl.maxclients) // we could still get unconnecteds!
+						{
+							// we parsed a valid ping entry, so expect another to follow
+							cl.parsingtextmode = CL_PARSETEXTMODE_PING;
+							cl.parsingtextexpectingpingforscores = expected;
+						}
+						return !expected;
+					}
+				}
+				if (!strncmp(t, "unconnected\n", 12))
+				{
+					// just ignore
+					cl.parsingtextmode = CL_PARSETEXTMODE_PING;
+					cl.parsingtextexpectingpingforscores = expected;
+					return !expected;
+				}
+				else
+					Con_DPrintf("player names '%s' and '%s' didn't match\n", cl.scores[cl.parsingtextplayerindex].name, t);
+			}
+		}
+	}
+	if (cl.parsingtextmode == CL_PARSETEXTMODE_STATUS)
+	{
+		if (!strncmp(text, "players: ", 9))
+		{
+			cl.parsingtextmode = CL_PARSETEXTMODE_STATUS_PLAYERID;
+			cl.parsingtextplayerindex = 0;
+			return true;
+		}
+		else if (!strstr(text, ": "))
+		{
+			cl.parsingtextmode = CL_PARSETEXTMODE_NONE; // status report ended
+			return true;
+		}
+	}
+	if (cl.parsingtextmode == CL_PARSETEXTMODE_STATUS_PLAYERID)
+	{
+		// if anything goes wrong, we'll assume this is not a status report
+		cl.parsingtextmode = CL_PARSETEXTMODE_NONE;
+		if (text[0] == '#' && text[1] >= '0' && text[1] <= '9')
+		{
+			t = text + 1;
+			cl.parsingtextplayerindex = atoi(t);
+			while (*t >= '0' && *t <= '9')
+				t++;
+			if (*t == ' ')
+			{
+				cl.parsingtextmode = CL_PARSETEXTMODE_STATUS_PLAYERIP;
+				return true;
+			}
+		}
+	}
+	if (cl.parsingtextmode == CL_PARSETEXTMODE_STATUS_PLAYERIP)
+	{
+		// if anything goes wrong, we'll assume this is not a status report
+		cl.parsingtextmode = CL_PARSETEXTMODE_NONE;
+		if (text[0] == ' ')
+		{
+			t = text;
+			while (*t == ' ')
+				t++;
+			// botclient is perfectly valid, but we don't care about bots
+			if (strcmp(t, "botclient\n"))
+			{
+				lhnetaddress_t address;
+				LHNETADDRESS_FromString(&address, t, 0);
+				// TODO: log the IP address and player name?
+			}
+			cl.parsingtextmode = CL_PARSETEXTMODE_STATUS_PLAYERID;
+			return true;
+		}
+	}
+	return true;
+}
+
 #define SHOWNET(x) if(cl_shownet.integer==2)Con_Printf("%3i:%s\n", msg_readcount-1, x);
 
 //[515]: csqc
@@ -2093,6 +2221,7 @@ void CL_ParseServerMessage(void)
 		cl.mtime[1] = cl.mtime[0];
 		cl.mtime[0] = realtime; // qw has no clock
 		cl.movement_needupdate = true;
+		cl.onground = false; // since there's no clientdata parsing, clear the onground flag here
 		// if true the cl.viewangles are interpolated from cl.mviewangles[]
 		// during this frame
 		// (makes spectating players much smoother and prevents mouse movement from turning)
@@ -2101,6 +2230,9 @@ void CL_ParseServerMessage(void)
 		if (!cls.demoplayback)
 			VectorCopy(cl.mviewangles[0], cl.mviewangles[1]);
 
+		// force a recalculation of the player prediction
+		cl.movement_replay = true;
+
 		// slightly kill qw player entities each frame
 		for (i = 1;i < cl.maxclients;i++)
 			cl.entities_active[i] = false;
@@ -2144,7 +2276,7 @@ void CL_ParseServerMessage(void)
 				{
 					char description[32*64], temp[64];
 					int count;
-					strcpy(description, "packet dump: ");
+					strlcpy(description, "packet dump: ", sizeof(description));
 					i = cmdcount - 32;
 					if (i < 0)
 						i = 0;
@@ -2178,10 +2310,14 @@ void CL_ParseServerMessage(void)
 
 			case qw_svc_print:
 				i = MSG_ReadByte();
-				if (i == 3) // chat
-					CSQC_AddPrintText(va("\1%s", MSG_ReadString()));	//[515]: csqc
-				else
-					CSQC_AddPrintText(MSG_ReadString());
+				temp = MSG_ReadString();
+				if (CL_ExaminePrintString(temp)) // look for anything interesting like player IP addresses or ping reports
+				{
+					if (i == 3) // chat
+						CSQC_AddPrintText(va("\1%s", temp));	//[515]: csqc
+					else
+						CSQC_AddPrintText(temp);
+				}
 				break;
 
 			case qw_svc_centerprint:
@@ -2263,7 +2399,7 @@ void CL_ParseServerMessage(void)
 				if (i >= cl.maxclients)
 					Host_Error("CL_ParseServerMessage: svc_updateentertime >= cl.maxclients");
 				// seconds ago
-				cl.scores[i].qw_entertime = realtime - MSG_ReadFloat();
+				cl.scores[i].qw_entertime = cl.time - MSG_ReadFloat();
 				break;
 
 			case qw_svc_spawnbaseline:
@@ -2480,7 +2616,7 @@ void CL_ParseServerMessage(void)
 				{
 					char description[32*64], temp[64];
 					int count;
-					strcpy (description, "packet dump: ");
+					strlcpy (description, "packet dump: ", sizeof(description));
 					i = cmdcount - 32;
 					if (i < 0)
 						i = 0;
@@ -2542,7 +2678,9 @@ void CL_ParseServerMessage(void)
 				break;
 
 			case svc_print:
-				CSQC_AddPrintText(MSG_ReadString());	//[515]: csqc
+				temp = MSG_ReadString();
+				if (CL_ExaminePrintString(temp)) // look for anything interesting like player IP addresses or ping reports
+					CSQC_AddPrintText(temp);	//[515]: csqc
 				break;
 
 			case svc_centerprint:
