@@ -199,6 +199,10 @@ float MSG_ReadAngle (protocolversion_t protocol);
 
 //============================================================================
 
+typedef float (*COM_CharWidthFunc_t) (void *passthrough, char c);
+typedef int (*COM_LineProcessorFunc) (void *passthrough, const char *line, size_t length, float width);
+int COM_Wordwrap(const char *string, size_t length, float maxWidth, COM_CharWidthFunc_t charWidth, void *passthroughCW, COM_LineProcessorFunc processLine, void *passthroughPL);
+
 extern char com_token[MAX_INPUTLINE];
 
 int COM_ParseToken(const char **datapointer, int returnnewline);
@@ -210,7 +214,6 @@ extern const char **com_argv;
 int COM_CheckParm (const char *parm);
 void COM_Init (void);
 void COM_Shutdown (void);
-void COM_InitArgv (void);
 void COM_InitGameType (void);
 
 char	*va(const char *format, ...);
@@ -228,6 +231,19 @@ char	*va(const char *format, ...);
 extern int dpsnprintf (char *buffer, size_t buffersize, const char *format, ...);
 extern int dpvsnprintf (char *buffer, size_t buffersize, const char *format, va_list args);
 
+// A bunch of functions are forbidden for security reasons (and also to please MSVS 2005, for some of them)
+// LordHavoc: added #undef lines here to avoid warnings in Linux
+#undef strcat
+#define strcat DO_NOT_USE_STRCAT__USE_STRLCAT_OR_MEMCPY
+#undef strncat
+#define strncat DO_NOT_USE_STRNCAT__USE_STRLCAT_OR_MEMCPY
+#undef strcpy
+#define strcpy DO_NOT_USE_STRCPY__USE_STRLCPY_OR_MEMCPY
+#undef strncpy
+#define strncpy DO_NOT_USE_STRNCPY__USE_STRLCPY_OR_MEMCPY
+//#undef sprintf
+//#define sprintf DO_NOT_USE_SPRINTF__USE_DPSNPRINTF
+
 
 //============================================================================
 
@@ -255,6 +271,8 @@ typedef enum gamemode_e
 	GAME_NETHERWORLD,
 	GAME_THEHUNTED,
 	GAME_DEFEATINDETAIL2,
+	GAME_DARSANA,
+	GAME_CONTAGIONTHEORY,
 }
 gamemode_t;
 
