@@ -25,30 +25,58 @@ Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 #include <time.h>
 #include "quakedef.h"
 
-int con_linewidth;
-
 float con_cursorspeed = 4;
 
 #define		CON_TEXTSIZE	131072
+#define		CON_MAXLINES	  4096
 
-// total lines in console scrollback
-int con_totallines;
 // lines up from bottom to display
 int con_backscroll;
-// where next message will be printed
-int con_current;
-// offset in current line for next print
-int con_x;
+
+// console buffer
 char con_text[CON_TEXTSIZE];
 
+#define CON_MASK_HIDENOTIFY 128
+#define CON_MASK_CHAT 1
+
+typedef struct
+{
+	char *start;
+	int len;
+
+	double addtime;
+	int mask;
+	
+	int height; // recalculated line height when needed (-1 to unset)
+}
+con_lineinfo;
+con_lineinfo con_lines[CON_MAXLINES];
+
+int con_lines_first; // cyclic buffer
+int con_lines_count;
+#define CON_LINES_IDX(i) ((con_lines_first + (i)) % CON_MAXLINES)
+#define CON_LINES_LAST CON_LINES_IDX(con_lines_count - 1)
+#define CON_LINES(i) con_lines[CON_LINES_IDX(i)]
+#define CON_LINES_PRED(i) (((i) + CON_MAXLINES - 1) % CON_MAXLINES)
+#define CON_LINES_SUCC(i) (((i) + 1) % CON_MAXLINES)
+
 cvar_t con_notifytime = {CVAR_SAVE, "con_notifytime","3", "how long notify lines last, in seconds"};
-cvar_t con_notify = {CVAR_SAVE, "con_notify","4", "how many notify lines to show (0-32)"};
+cvar_t con_notify = {CVAR_SAVE, "con_notify","4", "how many notify lines to show"};
+cvar_t con_chattime = {CVAR_SAVE, "con_chattime","30", "how long chat lines last, in seconds"};
+cvar_t con_chat = {CVAR_SAVE, "con_chat","5", "how many chat lines to show"};
+cvar_t con_chatwidth = {CVAR_SAVE, "con_chatwidth","1.0", "relative chat window width"};
 cvar_t con_textsize = {CVAR_SAVE, "con_textsize","8", "console text size in virtual 2D pixels"};	//[515]: console text size in pixels
 
-#define MAX_NOTIFYLINES 32
-// cl.time time the line was generated for transparent notify lines
-float con_times[MAX_NOTIFYLINES];
 
+cvar_t sys_specialcharactertranslation = {0, "sys_specialcharactertranslation", "1", "terminal console conchars to ASCII translation (set to 0 if your conchars.tga is for an 8bit character set or if you want raw output)"};
+#ifdef WIN32
+cvar_t sys_colortranslation = {0, "sys_colortranslation", "0",
+#else
+cvar_t sys_colortranslation = {0, "sys_colortranslation", "1",
+#endif
+                                                               "terminal console color translation (supported values: 0 = strip color codes, 1 = translate to ANSI codes, 2 = no translation)"};
+
+int con_linewidth;
 int con_vislines;
 
 qboolean con_initialized;
@@ -259,7 +287,7 @@ void Con_ToggleConsole_f (void)
 {
 	// toggle the 'user wants console' bit
 	key_consoleactive ^= KEY_CONSOLEACTIVE_USER;
-	memset (con_times, 0, sizeof(con_times));
+	Con_ClearNotify();
 }
 
 /*
@@ -269,22 +297,22 @@ Con_Clear_f
 */
 void Con_Clear_f (void)
 {
-	if (con_text)
-		memset (con_text, ' ', CON_TEXTSIZE);
+	con_lines_count = 0;
 }
 
 
 /*
 ================
 Con_ClearNotify
+
+Clear all notify lines.
 ================
 */
 void Con_ClearNotify (void)
 {
 	int i;
-
-	for (i=0 ; i<MAX_NOTIFYLINES ; i++)
-		con_times[i] = 0;
+	for(i = 0; i < con_lines_count; ++i)
+		CON_LINES(i).mask |= CON_MASK_HIDENOTIFY;
 }
 
 
@@ -321,9 +349,8 @@ If the line width has changed, reformat the buffer.
 */
 void Con_CheckResize (void)
 {
-	int i, j, width, oldwidth, oldtotallines, numlines, numchars;
+	int i, width, oldwidth;
 	float f;
-	char tbuf[CON_TEXTSIZE];
 
 	f = bound(1, con_textsize.value, 128);
 	if(f != con_textsize.value)
@@ -336,35 +363,12 @@ void Con_CheckResize (void)
 
 	oldwidth = con_linewidth;
 	con_linewidth = width;
-	oldtotallines = con_totallines;
-	con_totallines = CON_TEXTSIZE / con_linewidth;
-	numlines = oldtotallines;
 
-	if (con_totallines < numlines)
-		numlines = con_totallines;
-
-	numchars = oldwidth;
-
-	if (con_linewidth < numchars)
-		numchars = con_linewidth;
-
-	memcpy (tbuf, con_text, CON_TEXTSIZE);
-	memset (con_text, ' ', CON_TEXTSIZE);
-
-	for (i=0 ; i<numlines ; i++)
-	{
-		for (j=0 ; j<numchars ; j++)
-		{
-			con_text[(con_totallines - 1 - i) * con_linewidth + j] =
-					tbuf[((con_current - i + oldtotallines) %
-						  oldtotallines) * oldwidth + j];
-		}
-	}
-
-	Con_ClearNotify ();
+	for(i = 0; i < con_lines_count; ++i)
+		CON_LINES(i).height = -1; // recalculate when next needed
 
+	Con_ClearNotify();
 	con_backscroll = 0;
-	con_current = con_totallines - 1;
 }
 
 //[515]: the simplest command ever
@@ -382,6 +386,29 @@ static void Con_Maps_f (void)
 		GetMapList("", NULL, 0);
 }
 
+void Con_ConDump_f (void)
+{
+	int i;
+	qfile_t *file;
+	if (Cmd_Argc() != 2)
+	{
+		Con_Printf("usage: condump <filename>\n");
+		return;
+	}
+	file = FS_Open(Cmd_Argv(1), "wb", false, false);
+	if (!file)
+	{
+		Con_Printf("condump: unable to write file \"%s\"\n", file);
+		return;
+	}
+	for(i = 0; i < con_lines_count; ++i)
+	{
+		FS_Write(file, CON_LINES(i).start, CON_LINES(i).len);
+		FS_Write(file, "\n", 1);
+	}
+	FS_Close(file);
+}
+
 /*
 ================
 Con_Init
@@ -389,28 +416,31 @@ Con_Init
 */
 void Con_Init (void)
 {
-	memset (con_text, ' ', CON_TEXTSIZE);
 	con_linewidth = 80;
-	con_totallines = CON_TEXTSIZE / con_linewidth;
+	con_lines_first = 0;
+	con_lines_count = 0;
 
 	// Allocate a log queue, this will be freed after configs are parsed
 	logq_size = MAX_INPUTLINE;
 	logqueue = (unsigned char *)Mem_Alloc (tempmempool, logq_size);
 	logq_ind = 0;
 
+	Cvar_RegisterVariable (&sys_colortranslation);
+	Cvar_RegisterVariable (&sys_specialcharactertranslation);
+
 	Cvar_RegisterVariable (&log_file);
 
 	// support for the classic Quake option
 // COMMANDLINEOPTION: Console: -condebug logs console messages to qconsole.log, see also log_file
 	if (COM_CheckParm ("-condebug") != 0)
 		Cvar_SetQuick (&log_file, "qconsole.log");
-}
 
-void Con_Init_Commands (void)
-{
 	// register our cvars
 	Cvar_RegisterVariable (&con_notifytime);
 	Cvar_RegisterVariable (&con_notify);
+	Cvar_RegisterVariable (&con_chattime);
+	Cvar_RegisterVariable (&con_chat);
+	Cvar_RegisterVariable (&con_chatwidth);
 	Cvar_RegisterVariable (&con_textsize);
 
 	// register our commands
@@ -419,6 +449,7 @@ void Con_Init_Commands (void)
 	Cmd_AddCommand ("messagemode2", Con_MessageMode2_f, "input a chat message to say to only your team");
 	Cmd_AddCommand ("clear", Con_Clear_f, "clear console history");
 	Cmd_AddCommand ("maps", Con_Maps_f, "list information about available maps");	// By [515]
+	Cmd_AddCommand ("condump", Con_ConDump_f, "output console history to a file (see also log_file)");
 
 	con_initialized = true;
 	Con_Print("Console initialized.\n");
@@ -426,18 +457,106 @@ void Con_Init_Commands (void)
 
 
 /*
-===============
-Con_Linefeed
-===============
-*/
-void Con_Linefeed (void)
-{
-	if (con_backscroll)
-		con_backscroll++;
+================
+Con_DeleteLine
 
-	con_x = 0;
-	con_current++;
-	memset (&con_text[(con_current%con_totallines)*con_linewidth], ' ', con_linewidth);
+Deletes the first line from the console history.
+================
+*/
+void Con_DeleteLine()
+{
+	if(con_lines_count == 0)
+		return;
+	--con_lines_count;
+	con_lines_first = CON_LINES_IDX(1);
+}
+
+/*
+================
+Con_DeleteLastLine
+
+Deletes the last line from the console history.
+================
+*/
+void Con_DeleteLastLine()
+{
+	if(con_lines_count == 0)
+		return;
+	--con_lines_count;
+}
+
+/*
+================
+Con_BytesLeft
+
+Checks if there is space for a line of the given length, and if yes, returns a
+pointer to the start of such a space, and NULL otherwise.
+================
+*/
+char *Con_BytesLeft(int len)
+{
+	if(len > CON_TEXTSIZE)
+		return NULL;
+	if(con_lines_count == 0)
+		return con_text;
+	else
+	{
+		char *firstline_start = con_lines[con_lines_first].start;
+		char *lastline_onepastend = con_lines[CON_LINES_LAST].start + con_lines[CON_LINES_LAST].len;
+		// the buffer is cyclic, so we first have two cases...
+		if(firstline_start < lastline_onepastend) // buffer is contiguous
+		{
+			// put at end?
+			if(len <= con_text + CON_TEXTSIZE - lastline_onepastend)
+				return lastline_onepastend;
+			// put at beginning?
+			else if(len <= firstline_start - con_text)
+				return con_text;
+			else
+				return NULL;
+		}
+		else // buffer has a contiguous hole
+		{
+			if(len <= firstline_start - lastline_onepastend)
+				return lastline_onepastend;
+			else	
+				return NULL;
+		}
+	}
+}
+
+/*
+================
+Con_AddLine
+
+Appends a given string as a new line to the console.
+================
+*/
+void Con_AddLine(const char *line, int len, int mask)
+{
+	char *putpos;
+	con_lineinfo *p;
+	if(len >= CON_TEXTSIZE)
+	{
+		// line too large?
+		// only display end of line.
+		line += len - CON_TEXTSIZE + 1;
+		len = CON_TEXTSIZE - 1;
+	}
+	while(!(putpos = Con_BytesLeft(len + 1)) || con_lines_count >= CON_MAXLINES)
+		Con_DeleteLine();
+	memcpy(putpos, line, len);
+	putpos[len] = 0;
+	++con_lines_count;
+
+	//fprintf(stderr, "Now have %d lines (%d -> %d).\n", con_lines_count, con_lines_first, CON_LINES_LAST);
+
+	p = con_lines + CON_LINES_LAST;
+	p->start = putpos;
+	p->len = len;
+	p->addtime = cl.time;
+	p->mask = mask;
+	p->height = -1; // calculate when needed
 }
 
 /*
@@ -451,64 +570,43 @@ If no console is visible, the notify window will pop up.
 */
 void Con_PrintToHistory(const char *txt, int mask)
 {
-	int y, c, l;
-	static int cr;
+	// process:
+	//   \n goes to next line
+	//   \r deletes current line and makes a new one
 
-	while ( (c = *txt) )
+	static int cr_pending = 0;
+	static char buf[CON_TEXTSIZE];
+	static int bufpos = 0;
+
+	for(; *txt; ++txt)
 	{
-	// count word length
-		for (l=0 ; l< con_linewidth ; l++)
-			if ( txt[l] <= ' ')
+		if(cr_pending)
+		{
+			Con_DeleteLastLine();
+			cr_pending = 0;
+		}
+		switch(*txt)
+		{
+			case 0:
+				break;
+			case '\r':
+				Con_AddLine(buf, bufpos, mask);
+				bufpos = 0;
+				cr_pending = 1;
+				break;
+			case '\n':
+				Con_AddLine(buf, bufpos, mask);
+				bufpos = 0;
+				break;
+			default:
+				buf[bufpos++] = *txt;
+				if(bufpos >= CON_TEXTSIZE - 1)
+				{
+					Con_AddLine(buf, bufpos, mask);
+					bufpos = 0;
+				}
 				break;
-
-	// word wrap
-		if (l != con_linewidth && (con_x + l > con_linewidth) )
-			con_x = 0;
-
-		txt++;
-
-		if (cr)
-		{
-			con_current--;
-			cr = false;
 		}
-
-
-		if (!con_x)
-		{
-			Con_Linefeed ();
-		// mark time for transparent overlay
-			if (con_current >= 0)
-			{
-				if (con_notify.integer < 0)
-					Cvar_SetValueQuick(&con_notify, 0);
-				if (con_notify.integer > MAX_NOTIFYLINES)
-					Cvar_SetValueQuick(&con_notify, MAX_NOTIFYLINES);
-				if (con_notify.integer > 0)
-					con_times[con_current % con_notify.integer] = cl.time;
-			}
-		}
-
-		switch (c)
-		{
-		case '\n':
-			con_x = 0;
-			break;
-
-		case '\r':
-			con_x = 0;
-			cr = 1;
-			break;
-
-		default:	// display character and advance
-			y = con_current % con_totallines;
-			con_text[y*con_linewidth+con_x] = c | mask;
-			con_x++;
-			if (con_x >= con_linewidth)
-				con_x = 0;
-			break;
-		}
-
 	}
 }
 
@@ -561,7 +659,7 @@ extern cvar_t timeformat;
 extern qboolean sys_nostdout;
 void Con_Print(const char *msg)
 {
-	int mask = 0;
+	static int mask = 0;
 	static int index = 0;
 	static char line[MAX_INPUTLINE];
 
@@ -586,7 +684,10 @@ void Con_Print(const char *msg)
 			{
 				// play talk wav
 				if (*msg == 1)
+				{
 					S_LocalSound ("sound/misc/talk.wav");
+					mask = CON_MASK_CHAT;
+				}
 				line[index++] = STRING_COLOR_TAG;
 				line[index++] = '3';
 				msg++;
@@ -607,14 +708,162 @@ void Con_Print(const char *msg)
 			Log_ConPrint(line);
 			// send to scrollable buffer
 			if (con_initialized && cls.state != ca_dedicated)
+			{
 				Con_PrintToHistory(line, mask);
+				mask = 0;
+			}
 			// send to terminal or dedicated server window
 			if (!sys_nostdout)
 			{
 				unsigned char *p;
-				for (p = (unsigned char *) line;*p; p++)
-					*p = qfont_table[*p];
-				Sys_PrintToTerminal(line);
+				if(sys_specialcharactertranslation.integer)
+				{
+					for (p = (unsigned char *) line;*p; p++)
+						*p = qfont_table[*p];
+				}
+
+				if(sys_colortranslation.integer == 1) // ANSI
+				{
+					static char printline[MAX_INPUTLINE * 4 + 3];
+						// 2 can become 7 bytes, rounding that up to 8, and 3 bytes are added at the end
+						// a newline can transform into four bytes, but then prevents the three extra bytes from appearing
+					int lastcolor = 0;
+					const char *in;
+					char *out;
+					for(in = line, out = printline; *in; ++in)
+					{
+						switch(*in)
+						{
+							case STRING_COLOR_TAG:
+								switch(in[1])
+								{
+									case STRING_COLOR_TAG:
+										++in;
+										*out++ = STRING_COLOR_TAG;
+										break;
+									case '0':
+									case '7':
+										// normal color
+										++in;
+										if(lastcolor == 0) break; else lastcolor = 0;
+										*out++ = 0x1B; *out++ = '['; *out++ = 'm';
+										break;
+									case '1':
+										// light red
+										++in;
+										if(lastcolor == 1) break; else lastcolor = 1;
+										*out++ = 0x1B; *out++ = '['; *out++ = '1'; *out++ = ';'; *out++ = '3'; *out++ = '1'; *out++ = 'm';
+										break;
+									case '2':
+										// light green
+										++in;
+										if(lastcolor == 2) break; else lastcolor = 2;
+										*out++ = 0x1B; *out++ = '['; *out++ = '1'; *out++ = ';'; *out++ = '3'; *out++ = '2'; *out++ = 'm';
+										break;
+									case '3':
+										// yellow
+										++in;
+										if(lastcolor == 3) break; else lastcolor = 3;
+										*out++ = 0x1B; *out++ = '['; *out++ = '1'; *out++ = ';'; *out++ = '3'; *out++ = '3'; *out++ = 'm';
+										break;
+									case '4':
+										// light blue
+										++in;
+										if(lastcolor == 4) break; else lastcolor = 4;
+										*out++ = 0x1B; *out++ = '['; *out++ = '1'; *out++ = ';'; *out++ = '3'; *out++ = '4'; *out++ = 'm';
+										break;
+									case '5':
+										// light cyan
+										++in;
+										if(lastcolor == 5) break; else lastcolor = 5;
+										*out++ = 0x1B; *out++ = '['; *out++ = '1'; *out++ = ';'; *out++ = '3'; *out++ = '6'; *out++ = 'm';
+										break;
+									case '6':
+										// light magenta
+										++in;
+										if(lastcolor == 6) break; else lastcolor = 6;
+										*out++ = 0x1B; *out++ = '['; *out++ = '1'; *out++ = ';'; *out++ = '3'; *out++ = '5'; *out++ = 'm';
+										break;
+									// 7 handled above
+									case '8':
+									case '9':
+										// bold normal color
+										++in;
+										if(lastcolor == 8) break; else lastcolor = 8;
+										*out++ = 0x1B; *out++ = '['; *out++ = '0'; *out++ = ';'; *out++ = '1'; *out++ = 'm';
+										break;
+									default:
+										*out++ = STRING_COLOR_TAG;
+										break;
+								}
+								break;
+							case '\n':
+								if(lastcolor != 0)
+								{
+									*out++ = 0x1B; *out++ = '['; *out++ = 'm';
+									lastcolor = 0;
+								}
+								*out++ = *in;
+								break;
+							default:
+								*out++ = *in;
+								break;
+						}
+					}
+					if(lastcolor != 0)
+					{
+						*out++ = 0x1B;
+						*out++ = '[';
+						*out++ = 'm';
+					}
+					*out++ = 0;
+					Sys_PrintToTerminal(printline);
+				}
+				else if(sys_colortranslation.integer == 2) // Quake
+				{
+					Sys_PrintToTerminal(line);
+				}
+				else // strip
+				{
+					static char printline[MAX_INPUTLINE]; // it can only get shorter here
+					const char *in;
+					char *out;
+					for(in = line, out = printline; *in; ++in)
+					{
+						switch(*in)
+						{
+							case STRING_COLOR_TAG:
+								switch(in[1])
+								{
+									case STRING_COLOR_TAG:
+										++in;
+										*out++ = STRING_COLOR_TAG;
+										break;
+									case '0':
+									case '1':
+									case '2':
+									case '3':
+									case '4':
+									case '5':
+									case '6':
+									case '7':
+									case '8':
+									case '9':
+										++in;
+										break;
+									default:
+										*out++ = STRING_COLOR_TAG;
+										break;
+								}
+								break;
+							default:
+								*out++ = *in;
+								break;
+						}
+					}
+					*out++ = 0;
+					Sys_PrintToTerminal(printline);
+				}
 			}
 			// empty the line buffer
 			index = 0;
@@ -705,7 +954,8 @@ void Con_DrawInput (void)
 	if (!key_consoleactive)
 		return;		// don't draw anything
 
-	text = strcpy(editlinecopy, key_lines[edit_line]);
+	strlcpy(editlinecopy, key_lines[edit_line], sizeof(editlinecopy));
+	text = editlinecopy;
 
 	// Advanced Console Editing by Radix radix@planetquake.com
 	// Added/Modified by EvilTypeGuy eviltypeguy@qeradiant.com
@@ -735,6 +985,137 @@ void Con_DrawInput (void)
 //	key_lines[edit_line][key_linepos] = 0;
 }
 
+typedef struct
+{
+	float alignment; // 0 = left, 0.5 = center, 1 = right
+	float fontsize;
+	float x;
+	float y;
+	float width;
+
+	float ymin, ymax;
+
+	// PRIVATE:
+	int colorindex; // init to -1
+	char previous;
+	float previousWidth;
+}
+con_text_info_t;
+
+float Con_CharWidthFunc(void *passthrough, char c)
+{
+	con_text_info_t *ti = (con_text_info_t *) passthrough;
+	float w = ti->fontsize; // for proportional fonts, change THIS
+	char prev = ti->previous;
+	float prevW = ti->previousWidth;
+	ti->previous = c;
+	ti->previousWidth = w;
+
+	if(c == 0)
+	{
+		ti->colorindex = -1;
+		return 0;
+	}
+
+	if(prev == STRING_COLOR_TAG)
+		if(c >= '0' && c <= '9')
+			return -prevW;
+		else
+			return w;
+	else
+		return w;
+}
+
+int Con_CountLineFunc(void *passthrough, const char *line, size_t length, float width)
+{
+	(void) passthrough;
+	(void) line;
+	(void) length;
+	(void) width;
+	return 1;
+}
+
+int Con_DisplayLineFunc(void *passthrough, const char *line, size_t length, float width)
+{
+	con_text_info_t *ti = (con_text_info_t *) passthrough;
+
+	if(ti->y < ti->ymin - 0.001)
+		(void) 0;
+	else if(ti->y > ti->ymax - ti->fontsize + 0.001)
+		(void) 0;
+	else
+		DrawQ_ColoredString(ti->x + (ti->width - width) * ti->alignment, ti->y, line, length, ti->fontsize, ti->fontsize, 1.0, 1.0, 1.0, 1.0, 0, &(ti->colorindex));
+
+	ti->y += ti->fontsize;
+	return 1;
+}
+
+
+int Con_DrawNotifyRect(int mask_must, int mask_mustnot, float maxage, float x, float y, float width, float height, float fontsize, float alignment_x, float alignment_y)
+{
+	int i;
+	int lines = 0;
+	int maxlines = (int) floor(height / fontsize + 0.01f);
+	int startidx;
+	int nskip = 0;
+	double t = cl.time; // saved so it won't change
+	con_text_info_t ti;
+
+	ti.fontsize = fontsize;
+	ti.alignment = alignment_x;
+	ti.width = width;
+	ti.ymin = y;
+	ti.ymax = y + height;
+
+	// first find the first line to draw by backwards iterating and word wrapping to find their length...
+	startidx = con_lines_count;
+	for(i = con_lines_count - 1; i >= 0; --i)
+	{
+		con_lineinfo *l = &CON_LINES(i);
+		int mylines;
+
+		if((l->mask & mask_must) != mask_must)
+			continue;
+		if(l->mask & mask_mustnot)
+			continue;
+		if(maxage && (l->addtime < t - maxage))
+			continue;
+
+		// WE FOUND ONE!
+		// Calculate its actual height...
+		mylines = COM_Wordwrap(l->start, l->len, width, Con_CharWidthFunc, &ti, Con_CountLineFunc, &ti);
+		if(lines + mylines >= maxlines)
+		{
+			nskip = lines + mylines - maxlines;
+			lines = maxlines;
+			startidx = i;
+			break;
+		}
+		lines += mylines;
+		startidx = i;
+	}
+
+	// then center according to the calculated amount of lines...
+	ti.x = x;
+	ti.y = y + alignment_y * (height - lines * fontsize) - nskip * fontsize;
+
+	// then actually draw
+	for(i = startidx; i < con_lines_count; ++i)
+	{
+		con_lineinfo *l = &CON_LINES(i);
+
+		if((l->mask & mask_must) != mask_must)
+			continue;
+		if(l->mask & mask_mustnot)
+			continue;
+		if(maxage && (t - l->addtime) > maxage)
+			continue;
+
+		COM_Wordwrap(l->start, l->len, width, Con_CharWidthFunc, &ti, Con_DisplayLineFunc, &ti);
+	}
+
+	return lines;
+}
 
 /*
 ================
@@ -746,64 +1127,44 @@ Draws the last few lines of output transparently over the game top
 void Con_DrawNotify (void)
 {
 	float	x, v;
-	char	*text;
-	int		i, stop;
-	float	time;
+	static double lasttime = 0.0;
+	float align;
 	char	temptext[MAX_INPUTLINE];
-	int colorindex = -1; //-1 for default
+	int numChatlines;
+
+	if(cl.time < lasttime)
+	{
+		// fix up timers
+		int i;
+		for(i = 0; i < con_lines_count; ++i)
+			CON_LINES(i).addtime += cl.time - lasttime;
+	}
+	lasttime = cl.time;
+
+	numChatlines = con_chat.integer;
 
 	if (con_notify.integer < 0)
 		Cvar_SetValueQuick(&con_notify, 0);
-	if (con_notify.integer > MAX_NOTIFYLINES)
-		Cvar_SetValueQuick(&con_notify, MAX_NOTIFYLINES);
 	if (gamemode == GAME_TRANSFUSION)
-		v = 8;
+		v = 8; // vertical offset
 	else
 		v = 0;
-	// make a copy of con_current here so that we can't get in a runaway loop printing new messages while drawing the notify text
-	stop = con_current;
-	for (i= stop-con_notify.integer+1 ; i<=stop ; i++)
+	// GAME_NEXUIZ: center, otherwise left justify
+
+	align = 0;
+	if(gamemode == GAME_NEXUIZ)
+		align = 0.5;
+	v += con_textsize.value * Con_DrawNotifyRect(0, CON_MASK_HIDENOTIFY | (numChatlines ? CON_MASK_CHAT : 0), con_notifytime.value, 0, v, vid_conwidth.value, con_notify.value * con_textsize.value, con_textsize.value, align, 0.0);
+
+	// chat?
+	if(numChatlines)
 	{
-
-		if (i < 0)
-			continue;
-		time = con_times[i % con_notify.integer];
-		if (time == 0)
-			continue;
-		time = cl.time - time;
-		if (time > con_notifytime.value)
-			continue;
-		text = con_text + (i % con_totallines)*con_linewidth;
-
-		if (gamemode == GAME_NEXUIZ) {
-			int chars = 0;
-			int finalchars = 0;
-			int j;
-
-			// count up to the last non-whitespace, and ignore color codes
-			for (j = 0;j < con_linewidth && text[j];j++)
-			{
-				if (text[j] == '^' && (text[j+1] >= '0' && text[j+1] <= '9'))
-				{
-					j++;
-					continue;
-				}
-				chars++;
-				if (text[j] == ' ')
-					continue;
-				finalchars = chars;
-			}
-			// center the line using the calculated width
-			x = (vid_conwidth.integer - finalchars * con_textsize.value) * 0.5;
-		} else
-			x = 0;
-
-		DrawQ_ColoredString( x, v, text, con_linewidth, con_textsize.value, con_textsize.value, 1.0, 1.0, 1.0, 1.0, 0, &colorindex );
-
-		v += con_textsize.value;
+		float y1 = vid_conheight.value - 56;
+		float y0 = y1 - numChatlines * con_textsize.value;
+		Con_DrawNotifyRect(CON_MASK_CHAT, 0, con_chattime.value, 0, y0, vid_conwidth.value * con_chatwidth.value, y1 - y0, con_textsize.value, 0.0, 1.0);
+		v = y1;
 	}
 
-
 	if (key_dest == key_message)
 	{
 		int colorindex = -1;
@@ -818,7 +1179,7 @@ void Con_DrawNotify (void)
 		while ((int)strlen(temptext) >= con_linewidth)
 		{
 			DrawQ_ColoredString( 0, v, temptext, con_linewidth, con_textsize.value, con_textsize.value, 1.0, 1.0, 1.0, 1.0, 0, &colorindex );
-			strcpy(temptext, &temptext[con_linewidth]);
+			strlcpy(temptext, &temptext[con_linewidth], sizeof(temptext));
 			v += con_textsize.value;
 		}
 		if (strlen(temptext) > 0)
@@ -829,6 +1190,102 @@ void Con_DrawNotify (void)
 	}
 }
 
+/*
+================
+Con_MeasureConsoleLine
+
+Counts the number of lines for a line on the console.
+================
+*/
+int Con_MeasureConsoleLine(int lineno)
+{
+	float width = vid_conwidth.value;
+	con_text_info_t ti;
+	ti.fontsize = con_textsize.value;
+
+	return COM_Wordwrap(con_lines[lineno].start, con_lines[lineno].len, width, Con_CharWidthFunc, &ti, Con_CountLineFunc, NULL);
+}
+
+/*
+================
+Con_LineHeight
+
+Returns the height of a given console line; calculates it if necessary.
+================
+*/
+int Con_LineHeight(int i)
+{
+	int h = con_lines[i].height;
+	if(h != -1)
+		return h;
+	return con_lines[i].height = Con_MeasureConsoleLine(i);
+}
+
+/*
+================
+Con_DrawConsoleLine
+
+Draws a line of the console; returns its height in lines.
+If alpha is 0, the line is not drawn, but still wrapped and its height
+returned.
+================
+*/
+int Con_DrawConsoleLine(float y, int lineno, float ymin, float ymax)
+{
+	float width = vid_conwidth.value;
+
+	con_text_info_t ti;
+	ti.alignment = 0;
+	ti.fontsize = con_textsize.value;
+	ti.x = 0;
+	ti.y = y - (Con_LineHeight(lineno) - 1) * ti.fontsize;
+	ti.ymin = ymin;
+	ti.ymax = ymax;
+	ti.width = width;
+
+	return COM_Wordwrap(con_lines[lineno].start, con_lines[lineno].len, width, Con_CharWidthFunc, &ti, Con_DisplayLineFunc, &ti);
+}
+
+/*
+================
+Con_LastVisibleLine
+
+Calculates the last visible line index and how much to show of it based on
+con_backscroll.
+================
+*/
+void Con_LastVisibleLine(int *last, int *limitlast)
+{
+	int lines_seen = 0;
+	int ic;
+
+	if(con_backscroll < 0)
+		con_backscroll = 0;
+
+	// now count until we saw con_backscroll actual lines
+	for(ic = 0; ic < con_lines_count; ++ic)
+	{
+		int i = CON_LINES_IDX(con_lines_count - 1 - ic);
+		int h = Con_LineHeight(i);
+
+		// line is the last visible line?
+		if(lines_seen + h > con_backscroll && lines_seen <= con_backscroll)
+		{
+			*last = i;
+			*limitlast = lines_seen + h - con_backscroll;
+			return;
+		}
+
+		lines_seen += h;
+	}
+
+	// if we get here, no line was on screen - scroll so that one line is
+	// visible then.
+	con_backscroll = lines_seen - 1;
+	*last = con_lines_first;
+	*limitlast = 1;
+}
+
 /*
 ================
 Con_DrawConsole
@@ -839,32 +1296,39 @@ The typing input line at the bottom should only be drawn if typing is allowed
 */
 void Con_DrawConsole (int lines)
 {
-	int i, rows, j, stop;
+	int i, last, limitlast;
 	float y;
-	char *text;
-	int colorindex = -1;
 
 	if (lines <= 0)
 		return;
 
+	con_vislines = lines;
+
 // draw the background
 	DrawQ_Pic(0, lines - vid_conheight.integer, scr_conbrightness.value >= 0.01f ? Draw_CachePic("gfx/conback", true) : NULL, vid_conwidth.integer, vid_conheight.integer, scr_conbrightness.value, scr_conbrightness.value, scr_conbrightness.value, scr_conalpha.value, 0);
 	DrawQ_String(vid_conwidth.integer - strlen(engineversion) * con_textsize.value - con_textsize.value, lines - con_textsize.value, engineversion, 0, con_textsize.value, con_textsize.value, 1, 0, 0, 1, 0);
 
 // draw the text
-	con_vislines = lines;
-
-	rows = (int)ceil((lines/con_textsize.value)-2);		// rows of text to draw
-	y = lines - (rows+2)*con_textsize.value;	// may start slightly negative
-
-	// make a copy of con_current here so that we can't get in a runaway loop printing new messages while drawing the notify text
-	stop = con_current;
-	for (i = stop - rows + 1;i <= stop;i++, y += con_textsize.value)
+	if(con_lines_count > 0)
 	{
-		j = max(i - con_backscroll, 0);
-		text = con_text + (j % con_totallines)*con_linewidth;
+		float ymax = con_vislines - 2 * con_textsize.value;
+		Con_LastVisibleLine(&last, &limitlast);
+		y = ymax - con_textsize.value;
 
-		DrawQ_ColoredString( 0, y, text, con_linewidth, con_textsize.value, con_textsize.value, 1.0, 1.0, 1.0, 1.0, 0, &colorindex );
+		if(limitlast)
+			y += (con_lines[last].height - limitlast) * con_textsize.value;
+		i = last;
+
+		for(;;)
+		{
+			y -= Con_DrawConsoleLine(y, i, 0, ymax) * con_textsize.value;
+			if(i == con_lines_first)
+				break; // top of console buffer
+			if(y < 0)
+				break; // top of console window
+			limitlast = 0;
+			i = CON_LINES_PRED(i);
+		}
 	}
 
 // draw the input prompt, user text, and cursor if desired
@@ -918,7 +1382,7 @@ qboolean GetMapList (const char *s, char *completedname, int completednamebuffer
 		const char *data = NULL;
 		char keyname[64];
 		char entfilename[MAX_QPATH];
-		strcpy(message, "^1**ERROR**^7");
+		strlcpy(message, "^1**ERROR**^7", sizeof(message));
 		p = 0;
 		f = FS_Open(t->filenames[i], "rb", true, false);
 		if(f)
@@ -960,7 +1424,7 @@ qboolean GetMapList (const char *s, char *completedname, int completednamebuffer
 			else
 				p = 0;
 			strlcpy(entfilename, t->filenames[i], sizeof(entfilename));
-			strcpy(entfilename + strlen(entfilename) - 4, ".ent");
+			memcpy(entfilename + strlen(entfilename) - 4, ".ent", 5);
 			entities = (char *)FS_LoadFile(entfilename, tempmempool, true, NULL);
 			if (!entities && lumplen >= 10)
 			{
@@ -977,7 +1441,7 @@ qboolean GetMapList (const char *s, char *completedname, int completednamebuffer
 				for (;;)
 				{
 					int l;
-					if (!COM_ParseToken(&data, false))
+					if (!COM_ParseTokenConsole(&data))
 						break;
 					if (com_token[0] == '{')
 						continue;
@@ -988,7 +1452,7 @@ qboolean GetMapList (const char *s, char *completedname, int completednamebuffer
 					for (l = 0;l < (int)sizeof(keyname) - 1 && com_token[k+l] && com_token[k+l] > ' ';l++)
 						keyname[l] = com_token[k+l];
 					keyname[l] = 0;
-					if (!COM_ParseToken(&data, false))
+					if (!COM_ParseTokenConsole(&data))
 						break;
 					if (developer.integer >= 100)
 						Con_Printf("key: %s %s\n", keyname, com_token);
@@ -1008,12 +1472,12 @@ qboolean GetMapList (const char *s, char *completedname, int completednamebuffer
 		*(t->filenames[i]+len[i]+5) = 0;
 		switch(p)
 		{
-		case Q3BSPVERSION:	strcpy((char *)buf, "Q3");break;
-		case Q2BSPVERSION:	strcpy((char *)buf, "Q2");break;
-		case BSPVERSION:	strcpy((char *)buf, "Q1");break;
-		case MCBSPVERSION:	strcpy((char *)buf, "MC");break;
-		case 30:			strcpy((char *)buf, "HL");break;
-		default:			strcpy((char *)buf, "??");break;
+		case Q3BSPVERSION:	strlcpy((char *)buf, "Q3", sizeof(buf));break;
+		case Q2BSPVERSION:	strlcpy((char *)buf, "Q2", sizeof(buf));break;
+		case BSPVERSION:	strlcpy((char *)buf, "Q1", sizeof(buf));break;
+		case MCBSPVERSION:	strlcpy((char *)buf, "MC", sizeof(buf));break;
+		case 30:			strlcpy((char *)buf, "HL", sizeof(buf));break;
+		default:			strlcpy((char *)buf, "??", sizeof(buf));break;
 		}
 		Con_Printf("%16s (%s) %s\n", t->filenames[i]+5, buf, message);
 	}
@@ -1161,7 +1625,7 @@ void Con_CompleteCommandLine (void)
 	if (!(c + v + a))	// No possible matches
 	{
 		if(s2[0])
-			strcpy(&key_lines[edit_line][key_linepos], s2);
+			strlcpy(&key_lines[edit_line][key_linepos], s2, sizeof(key_lines[edit_line]) - key_linepos);
 		return;
 	}
 
