@@ -338,7 +338,7 @@ static void R_DrawPortal_Callback(const entity_render_t *ent, const rtlight_t *r
 	GL_BlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
 	GL_DepthMask(false);
 	GL_DepthTest(true);
-	qglDisable(GL_CULL_FACE);CHECKGLERROR
+	GL_CullFace(GL_NONE);
 	R_Mesh_Matrix(&identitymatrix);
 
 	numpoints = min(portal->numpoints, POLYGONELEMENTS_MAXPOINTS);
@@ -355,7 +355,6 @@ static void R_DrawPortal_Callback(const entity_render_t *ent, const rtlight_t *r
 	for (i = 0, v = vertex3f;i < numpoints;i++, v += 3)
 		VectorCopy(portal->points[i].position, v);
 	R_Mesh_Draw(0, numpoints, numpoints - 2, polygonelements);
-	qglEnable(GL_CULL_FACE);CHECKGLERROR
 }
 
 // LordHavoc: this is just a nice debugging tool, very slow
@@ -771,7 +770,7 @@ static void R_Q1BSP_DrawLight_TransparentCallback(const entity_render_t *ent, co
 	int surfacelistindex, batchcount;
 	texture_t *t;
 	msurface_t *batchsurfaces[BATCHSIZE];
-	// note: in practice this never actualy batches, oh well
+	// note: in practice this never actually batches, oh well
 	R_Shadow_RenderMode_Begin();
 	R_Shadow_RenderMode_ActiveLight((rtlight_t *)rtlight);
 	R_Shadow_RenderMode_Lighting(false, true);
@@ -874,7 +873,6 @@ void R_Q1BSP_DrawLight(entity_render_t *ent, int numsurfaces, const int *surface
 			R_Shadow_RenderSurfacesLighting(batchnumsurfaces, batchsurfacelist);
 		batchnumsurfaces = 0;
 	}
-	qglEnable(GL_CULL_FACE);CHECKGLERROR
 }
 
 //Made by [515]
@@ -905,7 +903,7 @@ void R_ReplaceWorldTexture (void)
 	{
 		if(t->width && !strcasecmp(t->name, r))
 		{
-			if(Mod_LoadSkinFrame(&t->skin, (char*)newt, TEXF_MIPMAP | TEXF_ALPHA | TEXF_PRECACHE | TEXF_PICMIP, false, r_fullbrights.integer))
+			if(Mod_LoadSkinFrame(&t->skinframes[0], (char*)newt, TEXF_MIPMAP | TEXF_ALPHA | TEXF_PRECACHE | TEXF_PICMIP, false, r_fullbrights.integer))
 			{
 				Con_Printf("%s replaced with %s\n", r, newt);
 				return;
@@ -913,7 +911,7 @@ void R_ReplaceWorldTexture (void)
 			else
 			{
 				Con_Printf("%s was not found\n", newt);
-				Mod_LoadSkinFrame(&t->skin, (char*)r, TEXF_MIPMAP | TEXF_ALPHA | TEXF_PRECACHE | TEXF_PICMIP, false, r_fullbrights.integer);//back to default
+				Mod_LoadSkinFrame(&t->skinframes[0], (char*)r, TEXF_MIPMAP | TEXF_ALPHA | TEXF_PRECACHE | TEXF_PICMIP, false, r_fullbrights.integer);//back to default
 				return;
 			}
 		}
@@ -930,7 +928,7 @@ void R_ListWorldTextures (void)
 
 	Con_Print("Worldmodel textures :\n");
 	for(i=0,t=m->data_textures;i<m->num_textures;i++,t++)
-		if(t->skin.base != r_texture_notexture)
+		if (t->numskinframes)
 			Con_Printf("%s\n", t->name);
 }
 
