@@ -110,6 +110,7 @@ typedef double GLclampd;
 #define GL_TEXTURE_BORDER_COLOR			0x1004
 #define GL_TEXTURE_MAG_FILTER			0x2800
 #define GL_TEXTURE_MIN_FILTER			0x2801
+#define GL_PACK_ALIGNMENT			0x0D05
 #define GL_UNPACK_ALIGNMENT			0x0CF5
 #define GL_TEXTURE_BINDING_1D                   0x8068
 #define GL_TEXTURE_BINDING_2D                   0x8069
@@ -169,8 +170,20 @@ extern int gl_max_anisotropy;
 #define GL_TEXTURE_COORD_ARRAY			0x8078
 //#define GL_EDGE_FLAG_ARRAY			0x8079
 
+#define GL_NONE					0
+#define GL_FRONT_LEFT			0x0400
+#define GL_FRONT_RIGHT			0x0401
+#define GL_BACK_LEFT			0x0402
+#define GL_BACK_RIGHT			0x0403
 #define GL_FRONT				0x0404
 #define GL_BACK					0x0405
+#define GL_LEFT					0x0406
+#define GL_RIGHT				0x0407
+#define GL_FRONT_AND_BACK		0x0408
+#define GL_AUX0					0x0409
+#define GL_AUX1					0x040A
+#define GL_AUX2					0x040B
+#define GL_AUX3					0x040C
 
 #define GL_VENDOR				0x1F00
 #define GL_RENDERER				0x1F01
@@ -226,6 +239,9 @@ extern int gl_max_anisotropy;
 #define GL_POINT_SMOOTH                         0x0B10
 #define GL_LINE_SMOOTH                          0x0B20
 #define GL_POLYGON_SMOOTH                       0x0B41
+
+#define GL_POLYGON_STIPPLE                0x0B42
+
 #endif
 
 extern int gl_max_texture_size;
@@ -540,6 +556,7 @@ extern void (GLAPIENTRY *qglGetActiveUniformARB)(GLhandleARB programObj, GLuint
 extern void (GLAPIENTRY *qglGetUniformfvARB)(GLhandleARB programObj, GLint location, GLfloat *params);
 extern void (GLAPIENTRY *qglGetUniformivARB)(GLhandleARB programObj, GLint location, GLint *params);
 extern void (GLAPIENTRY *qglGetShaderSourceARB)(GLhandleARB obj, GLsizei maxLength, GLsizei *length, GLcharARB *source);
+extern void (GLAPIENTRY *qglPolygonStipple)(const GLubyte *mask);
 #ifndef GL_PROGRAM_OBJECT_ARB
 #define GL_PROGRAM_OBJECT_ARB					0x8B40
 #define GL_OBJECT_TYPE_ARB						0x8B4E
