@@ -71,8 +71,8 @@ sizebuf_t net_message;
 static unsigned char net_message_buf[NET_MAXMESSAGE];
 
 cvar_t net_messagetimeout = {0, "net_messagetimeout","300", "drops players who have not sent any packets for this many seconds"};
-cvar_t net_messagerejointimeout = {0, "net_messagerejointimeout","10", "give a player this much time in seconds to rejoin and continue playing (not losing frags and such)"};
 cvar_t net_connecttimeout = {0, "net_connecttimeout","10", "after requesting a connection, the client must reply within this many seconds or be dropped (cuts down on connect floods)"};
+cvar_t net_connectfloodblockingtimeout = {0, "net_connectfloodblockingtimeout", "5", "when a connection packet is received, it will block all future connect packets from that IP address for this many seconds (cuts down on connect floods)"};
 cvar_t hostname = {CVAR_SAVE, "hostname", "UNNAMED", "server message to show in server browser"};
 cvar_t developer_networking = {0, "developer_networking", "0", "prints all received and sent packets (recommended only for debugging)"};
 
@@ -787,10 +787,15 @@ netconn_t *NetConn_Open(lhnetsocket_t *mysocket, lhnetaddress_t *peeraddress)
 	return conn;
 }
 
+void NetConn_ClearConnectFlood(lhnetaddress_t *peeraddress);
 void NetConn_Close(netconn_t *conn)
 {
 	netconn_t *c;
 	// remove connection from list
+
+	// allow the client to reconnect immediately
+	NetConn_ClearConnectFlood(&(conn->peeraddress));
+
 	if (conn == netconn_list)
 		netconn_list = conn->next;
 	else
@@ -841,11 +846,13 @@ void NetConn_UpdateSockets(void)
 	}
 }
 
-static int NetConn_ReceivedMessage(netconn_t *conn, unsigned char *data, int length, protocolversion_t protocol)
+static int NetConn_ReceivedMessage(netconn_t *conn, unsigned char *data, int length, protocolversion_t protocol, double newtimeout)
 {
 	if (length < 8)
 		return 0;
 
+	// TODO: add netgraph stuff rather than just packetloss counting...
+
 	if (protocol == PROTOCOL_QUAKEWORLD)
 	{
 		int sequence, sequence_ack;
@@ -876,15 +883,22 @@ static int NetConn_ReceivedMessage(netconn_t *conn, unsigned char *data, int len
 		sequence_ack &= ~(1<<31);
 		if (sequence <= conn->qw.incoming_sequence)
 		{
-			Con_DPrint("Got a stale datagram\n");
+			//Con_DPrint("Got a stale datagram\n");
 			return 0;
 		}
 		count = sequence - (conn->qw.incoming_sequence + 1);
 		if (count > 0)
 		{
 			droppedDatagrams += count;
-			Con_DPrintf("Dropped %u datagram(s)\n", count);
+			//Con_DPrintf("Dropped %u datagram(s)\n", count);
+			while (count--)
+			{
+				conn->packetlost[conn->packetlostcounter] = true;
+				conn->packetlostcounter = (conn->packetlostcounter + 1) % 100;
+			}
 		}
+		conn->packetlost[conn->packetlostcounter] = false;
+		conn->packetlostcounter = (conn->packetlostcounter + 1) % 100;
 		if (reliable_ack == conn->qw.reliable_sequence)
 		{
 			// received, now we will be able to send another reliable message
@@ -899,7 +913,7 @@ static int NetConn_ReceivedMessage(netconn_t *conn, unsigned char *data, int len
 		if (reliable_message)
 			conn->qw.incoming_reliable_sequence ^= 1;
 		conn->lastMessageTime = realtime;
-		conn->timeout = realtime + net_messagetimeout.value;
+		conn->timeout = realtime + newtimeout;
 		unreliableMessagesReceived++;
 		SZ_Clear(&net_message);
 		SZ_Write(&net_message, data, length);
@@ -931,11 +945,18 @@ static int NetConn_ReceivedMessage(netconn_t *conn, unsigned char *data, int len
 					{
 						count = sequence - conn->nq.unreliableReceiveSequence;
 						droppedDatagrams += count;
-						Con_DPrintf("Dropped %u datagram(s)\n", count);
+						//Con_DPrintf("Dropped %u datagram(s)\n", count);
+						while (count--)
+						{
+							conn->packetlost[conn->packetlostcounter] = true;
+							conn->packetlostcounter = (conn->packetlostcounter + 1) % 100;
+						}
 					}
+					conn->packetlost[conn->packetlostcounter] = false;
+					conn->packetlostcounter = (conn->packetlostcounter + 1) % 100;
 					conn->nq.unreliableReceiveSequence = sequence + 1;
 					conn->lastMessageTime = realtime;
-					conn->timeout = realtime + net_messagetimeout.value;
+					conn->timeout = realtime + newtimeout;
 					unreliableMessagesReceived++;
 					if (length > 0)
 					{
@@ -945,8 +966,8 @@ static int NetConn_ReceivedMessage(netconn_t *conn, unsigned char *data, int len
 						return 2;
 					}
 				}
-				else
-					Con_DPrint("Got a stale datagram\n");
+				//else
+				//	Con_DPrint("Got a stale datagram\n");
 				return 1;
 			}
 			else if (flags & NETFLAG_ACK)
@@ -959,7 +980,7 @@ static int NetConn_ReceivedMessage(netconn_t *conn, unsigned char *data, int len
 						if (conn->nq.ackSequence != conn->nq.sendSequence)
 							Con_DPrint("ack sequencing error\n");
 						conn->lastMessageTime = realtime;
-						conn->timeout = realtime + net_messagetimeout.value;
+						conn->timeout = realtime + newtimeout;
 						if (conn->sendMessageLength > MAX_PACKETFRAGMENT)
 						{
 							unsigned int packetLen;
@@ -999,11 +1020,11 @@ static int NetConn_ReceivedMessage(netconn_t *conn, unsigned char *data, int len
 						else
 							conn->sendMessageLength = 0;
 					}
-					else
-						Con_DPrint("Duplicate ACK received\n");
+					//else
+					//	Con_DPrint("Duplicate ACK received\n");
 				}
-				else
-					Con_DPrint("Stale ACK received\n");
+				//else
+				//	Con_DPrint("Stale ACK received\n");
 				return 1;
 			}
 			else if (flags & NETFLAG_DATA)
@@ -1015,7 +1036,7 @@ static int NetConn_ReceivedMessage(netconn_t *conn, unsigned char *data, int len
 				if (sequence == conn->nq.receiveSequence)
 				{
 					conn->lastMessageTime = realtime;
-					conn->timeout = realtime + net_messagetimeout.value;
+					conn->timeout = realtime + newtimeout;
 					conn->nq.receiveSequence++;
 					if( conn->receiveMessageLength + length <= (int)sizeof( conn->receiveMessage ) ) {
 						memcpy(conn->receiveMessage + conn->receiveMessageLength, data, length);
@@ -1069,6 +1090,18 @@ void NetConn_ConnectionEstablished(lhnetsocket_t *mysocket, lhnetaddress_t *peer
 	cls.protocol = initialprotocol;
 	if (cls.protocol == PROTOCOL_QUAKEWORLD)
 		Cmd_ForwardStringToServer("new");
+	if (cls.protocol == PROTOCOL_QUAKE)
+	{
+		// write a keepalive (svc_nop) as it seems to greatly improve the
+		// chances of connecting to a netquake server
+		sizebuf_t msg;
+		unsigned char buf[4];
+		memset(&msg, 0, sizeof(msg));
+		msg.data = buf;
+		msg.maxsize = sizeof(buf);
+		MSG_WriteChar(&msg, svc_nop);
+		NetConn_SendUnreliableMessage(cls.netcon, &msg, cls.protocol);
+	}
 }
 
 int NetConn_IsLocalGame(void)
@@ -1394,7 +1427,7 @@ static int NetConn_ClientParsePacket(lhnetsocket_t *mysocket, unsigned char *dat
 		return true;
 	}
 	// quakeworld ingame packet
-	if (fromserver && cls.protocol == PROTOCOL_QUAKEWORLD && length >= 8 && (ret = NetConn_ReceivedMessage(cls.netcon, data, length, cls.protocol)) == 2)
+	if (fromserver && cls.protocol == PROTOCOL_QUAKEWORLD && length >= 8 && (ret = NetConn_ReceivedMessage(cls.netcon, data, length, cls.protocol, net_messagetimeout.value)) == 2)
 	{
 		ret = 0;
 		CL_ParseServerMessage();
@@ -1425,9 +1458,9 @@ static int NetConn_ClientParsePacket(lhnetsocket_t *mysocket, unsigned char *dat
 					length -= 4;
 					LHNETADDRESS_SetPort(&clientportaddress, port);
 				}
-				M_Update_Return_Reason("Accepted");
 				// update the server IP in the userinfo (QW servers expect this, and it is used by the reconnect command)
 				InfoString_SetValue(cls.userinfo, sizeof(cls.userinfo), "*ip", addressstring2);
+				M_Update_Return_Reason("Accepted");
 				NetConn_ConnectionEstablished(mysocket, &clientportaddress, PROTOCOL_QUAKE);
 			}
 			break;
@@ -1437,7 +1470,6 @@ static int NetConn_ClientParsePacket(lhnetsocket_t *mysocket, unsigned char *dat
 			cls.connect_trying = false;
 			M_Update_Return_Reason((char *)data);
 			break;
-		// TODO: fix this code so that lan searches for quake servers will work
 		case CCREP_SERVER_INFO:
 			if (developer.integer >= 10)
 				Con_Printf("Datagram_ParseConnectionless: received CCREP_SERVER_INFO from %s.\n", addressstring2);
@@ -1479,7 +1511,7 @@ static int NetConn_ClientParsePacket(lhnetsocket_t *mysocket, unsigned char *dat
 		return true;
 	}
 	ret = 0;
-	if (fromserver && length >= (int)NET_HEADERSIZE && (ret = NetConn_ReceivedMessage(cls.netcon, data, length, cls.protocol)) == 2)
+	if (fromserver && length >= (int)NET_HEADERSIZE && (ret = NetConn_ReceivedMessage(cls.netcon, data, length, cls.protocol, net_messagetimeout.value)) == 2)
 		CL_ParseServerMessage();
 	return ret;
 }
@@ -1668,7 +1700,7 @@ static qboolean NetConn_BuildStatusResponse(const char* challenge, char* out_msg
 			client_t *cl = &svs.clients[i];
 			if (cl->active)
 			{
-				int nameind, cleanind;
+				int nameind, cleanind, pingvalue;
 				char curchar;
 				char cleanname [sizeof(cl->name)];
 
@@ -1686,9 +1718,14 @@ static qboolean NetConn_BuildStatusResponse(const char* challenge, char* out_msg
 					}
 				} while (curchar != '\0');
 
+				pingvalue = (int)(cl->ping * 1000.0f);
+				if(cl->netconnection)
+					pingvalue = bound(1, pingvalue, 9999);
+				else
+					pingvalue = 0;
 				length = dpsnprintf(ptr, left, "%d %d \"%s\"\n",
 									cl->frags,
-									(int)(cl->ping * 1000.0f),
+									pingvalue,
 									cleanname);
 				if(length < 0)
 					return false;
@@ -1701,13 +1738,72 @@ static qboolean NetConn_BuildStatusResponse(const char* challenge, char* out_msg
 	return true;
 }
 
+static qboolean NetConn_PreventConnectFlood(lhnetaddress_t *peeraddress)
+{
+	int floodslotnum, bestfloodslotnum;
+	double bestfloodtime;
+	lhnetaddress_t noportpeeraddress;
+	// see if this is a connect flood
+	noportpeeraddress = *peeraddress;
+	LHNETADDRESS_SetPort(&noportpeeraddress, 0);
+	bestfloodslotnum = 0;
+	bestfloodtime = sv.connectfloodaddresses[bestfloodslotnum].lasttime;
+	for (floodslotnum = 0;floodslotnum < MAX_CONNECTFLOODADDRESSES;floodslotnum++)
+	{
+		if (bestfloodtime >= sv.connectfloodaddresses[floodslotnum].lasttime)
+		{
+			bestfloodtime = sv.connectfloodaddresses[floodslotnum].lasttime;
+			bestfloodslotnum = floodslotnum;
+		}
+		if (sv.connectfloodaddresses[floodslotnum].lasttime && LHNETADDRESS_Compare(&noportpeeraddress, &sv.connectfloodaddresses[floodslotnum].address) == 0)
+		{
+			// this address matches an ongoing flood address
+			if (realtime < sv.connectfloodaddresses[floodslotnum].lasttime + net_connectfloodblockingtimeout.value)
+			{
+				// renew the ban on this address so it does not expire
+				// until the flood has subsided
+				sv.connectfloodaddresses[floodslotnum].lasttime = realtime;
+				//Con_Printf("Flood detected!\n");
+				return true;
+			}
+			// the flood appears to have subsided, so allow this
+			bestfloodslotnum = floodslotnum; // reuse the same slot
+			break;
+		}
+	}
+	// begin a new timeout on this address
+	sv.connectfloodaddresses[bestfloodslotnum].address = noportpeeraddress;
+	sv.connectfloodaddresses[bestfloodslotnum].lasttime = realtime;
+	//Con_Printf("Flood detection initiated!\n");
+	return false;
+}
+
+void NetConn_ClearConnectFlood(lhnetaddress_t *peeraddress)
+{
+	int floodslotnum;
+	lhnetaddress_t noportpeeraddress;
+	// see if this is a connect flood
+	noportpeeraddress = *peeraddress;
+	LHNETADDRESS_SetPort(&noportpeeraddress, 0);
+	for (floodslotnum = 0;floodslotnum < MAX_CONNECTFLOODADDRESSES;floodslotnum++)
+	{
+		if (sv.connectfloodaddresses[floodslotnum].lasttime && LHNETADDRESS_Compare(&noportpeeraddress, &sv.connectfloodaddresses[floodslotnum].address) == 0)
+		{
+			// this address matches an ongoing flood address
+			// remove the ban
+			sv.connectfloodaddresses[floodslotnum].address.addresstype = LHNETADDRESSTYPE_NONE;
+			sv.connectfloodaddresses[floodslotnum].lasttime = 0;
+			//Con_Printf("Flood cleared!\n");
+		}
+	}
+}
+
 extern void SV_SendServerinfo (client_t *client);
 static int NetConn_ServerParsePacket(lhnetsocket_t *mysocket, unsigned char *data, int length, lhnetaddress_t *peeraddress)
 {
 	int i, ret, clientnum, best;
 	double besttime;
 	client_t *client;
-	netconn_t *conn;
 	char *s, *string, response[1400], addressstring2[128], stringbuf[16384];
 
 	if (!sv.active)
@@ -1768,81 +1864,83 @@ static int NetConn_ServerParsePacket(lhnetsocket_t *mysocket, unsigned char *dat
 		{
 			string += 7;
 			length -= 7;
-			if ((s = SearchInfostring(string, "challenge")))
+
+			if (!(s = SearchInfostring(string, "challenge")))
+				return true;
+			// validate the challenge
+			for (i = 0;i < MAX_CHALLENGES;i++)
+				if (!LHNETADDRESS_Compare(peeraddress, &challenge[i].address) && !strcmp(challenge[i].string, s))
+					break;
+			// if the challenge is not recognized, drop the packet
+			if (i == MAX_CHALLENGES)
+				return true;
+
+			// check engine protocol
+			if (strcmp(SearchInfostring(string, "protocol"), "darkplaces 3"))
 			{
-				// validate the challenge
-				for (i = 0;i < MAX_CHALLENGES;i++)
-					if (!LHNETADDRESS_Compare(peeraddress, &challenge[i].address) && !strcmp(challenge[i].string, s))
-						break;
-				if (i < MAX_CHALLENGES)
+				if (developer.integer >= 10)
+					Con_Printf("Datagram_ParseConnectionless: sending \"reject Wrong game protocol.\" to %s.\n", addressstring2);
+				NetConn_WriteString(mysocket, "\377\377\377\377reject Wrong game protocol.", peeraddress);
+				return true;
+			}
+
+			// see if this is a duplicate connection request or a disconnected
+			// client who is rejoining to the same client slot
+			for (clientnum = 0, client = svs.clients;clientnum < svs.maxclients;clientnum++, client++)
+			{
+				if (client->netconnection && LHNETADDRESS_Compare(peeraddress, &client->netconnection->peeraddress) == 0)
 				{
-					// check engine protocol
-					if (strcmp(SearchInfostring(string, "protocol"), "darkplaces 3"))
+					// this is a known client...
+					if (client->spawned)
 					{
+						// client crashed and is coming back,
+						// keep their stuff intact
 						if (developer.integer >= 10)
-							Con_Printf("Datagram_ParseConnectionless: sending \"reject Wrong game protocol.\" to %s.\n", addressstring2);
-						NetConn_WriteString(mysocket, "\377\377\377\377reject Wrong game protocol.", peeraddress);
+							Con_Printf("Datagram_ParseConnectionless: sending \"accept\" to %s.\n", addressstring2);
+						NetConn_WriteString(mysocket, "\377\377\377\377accept", peeraddress);
+						SV_VM_Begin();
+						SV_SendServerinfo(client);
+						SV_VM_End();
 					}
 					else
 					{
-						// see if this is a duplicate connection request
-						for (clientnum = 0, client = svs.clients;clientnum < svs.maxclients;clientnum++, client++)
-							if (client->netconnection && LHNETADDRESS_Compare(peeraddress, &client->netconnection->peeraddress) == 0)
-								break;
-						if (clientnum < svs.maxclients && realtime - client->connecttime < net_messagerejointimeout.value)
-						{
-							// client is still trying to connect,
-							// so we send a duplicate reply
-							if (developer.integer >= 10)
-								Con_Printf("Datagram_ParseConnectionless: sending duplicate accept to %s.\n", addressstring2);
-							NetConn_WriteString(mysocket, "\377\377\377\377accept", peeraddress);
-						}
-#if 0
-						else if (clientnum < svs.maxclients)
-						{
-							if (realtime - client->netconnection->lastMessageTime >= net_messagerejointimeout.value)
-							{
-								// client crashed and is coming back, keep their stuff intact
-								SV_SendServerinfo(client);
-								//host_client = client;
-								//SV_DropClient (true);
-							}
-							// else ignore them
-						}
-#endif
-						else
-						{
-							// this is a new client, find a slot
-							for (clientnum = 0, client = svs.clients;clientnum < svs.maxclients;clientnum++, client++)
-								if (!client->active)
-									break;
-							if (clientnum < svs.maxclients)
-							{
-								// prepare the client struct
-								if ((conn = NetConn_Open(mysocket, peeraddress)))
-								{
-									// allocated connection
-									if (developer.integer >= 10)
-										Con_Printf("Datagram_ParseConnectionless: sending \"accept\" to %s.\n", conn->address);
-									NetConn_WriteString(mysocket, "\377\377\377\377accept", peeraddress);
-									// now set up the client
-									SV_VM_Begin();
-									SV_ConnectClient(clientnum, conn);
-									SV_VM_End();
-									NetConn_Heartbeat(1);
-								}
-							}
-							else
-							{
-								// server is full
-								if (developer.integer >= 10)
-									Con_Printf("Datagram_ParseConnectionless: sending \"reject Server is full.\" to %s.\n", addressstring2);
-								NetConn_WriteString(mysocket, "\377\377\377\377reject Server is full.", peeraddress);
-							}
-						}
+						// client is still trying to connect,
+						// so we send a duplicate reply
+						if (developer.integer >= 10)
+							Con_Printf("Datagram_ParseConnectionless: sending duplicate accept to %s.\n", addressstring2);
+						NetConn_WriteString(mysocket, "\377\377\377\377accept", peeraddress);
 					}
+					return true;
 				}
 			}
+
+			if (NetConn_PreventConnectFlood(peeraddress))
+				return true;
+
+			// find an empty client slot for this new client
+			for (clientnum = 0, client = svs.clients;clientnum < svs.maxclients;clientnum++, client++)
+			{
+				netconn_t *conn;
+				if (!client->active && (conn = NetConn_Open(mysocket, peeraddress)))
+				{
+					// allocated connection
+					if (developer.integer >= 10)
+						Con_Printf("Datagram_ParseConnectionless: sending \"accept\" to %s.\n", conn->address);
+					NetConn_WriteString(mysocket, "\377\377\377\377accept", peeraddress);
+					// now set up the client
+					SV_VM_Begin();
+					SV_ConnectClient(clientnum, conn);
+					SV_VM_End();
+					NetConn_Heartbeat(1);
+					return true;
+				}
+			}
+
+			// no empty slots found - server is full
+			if (developer.integer >= 10)
+				Con_Printf("Datagram_ParseConnectionless: sending \"reject Server is full.\" to %s.\n", addressstring2);
+			NetConn_WriteString(mysocket, "\377\377\377\377reject Server is full.", peeraddress);
+
 			return true;
 		}
 		if (length >= 7 && !memcmp(string, "getinfo", 7))
@@ -1930,213 +2028,214 @@ static int NetConn_ServerParsePacket(lhnetsocket_t *mysocket, unsigned char *dat
 		// we're done processing this packet now
 		return true;
 	}
-	// LordHavoc: disabled netquake control packet support in server
-#if 0
+	// netquake control packets, supported for compatibility only, and only
+	// when running game protocols that are normally served via this connection
+	// protocol
+	// (this protects more modern protocols against being used for
+	//  Quake packet flood Denial Of Service attacks)
+	if (length >= 5 && (i = BigLong(*((int *)data))) && (i & (~NETFLAG_LENGTH_MASK)) == (int)NETFLAG_CTL && (i & NETFLAG_LENGTH_MASK) == length && (sv.protocol == PROTOCOL_QUAKE || sv.protocol == PROTOCOL_QUAKEDP || sv.protocol == PROTOCOL_NEHAHRAMOVIE || sv.protocol == PROTOCOL_DARKPLACES1 || sv.protocol == PROTOCOL_DARKPLACES2 || sv.protocol == PROTOCOL_DARKPLACES3))
 	{
-		int c, control;
-		// netquake control packets, supported for compatibility only
-		if (length >= 5 && (control = BigLong(*((int *)data))) && (control & (~NETFLAG_LENGTH_MASK)) == (int)NETFLAG_CTL && (control & NETFLAG_LENGTH_MASK) == length)
+		int c = data[4];
+		data += 5;
+		length -= 5;
+		switch (c)
 		{
-			c = data[4];
-			data += 5;
-			length -= 5;
-			switch (c)
+		case CCREQ_CONNECT:
+			if (developer.integer >= 10)
+				Con_Printf("Datagram_ParseConnectionless: received CCREQ_CONNECT from %s.\n", addressstring2);
+			if (length < (int)strlen("QUAKE") + 1 + 1)
+				break;
+
+			if (memcmp(data, "QUAKE", strlen("QUAKE") + 1) != 0 || (int)data[strlen("QUAKE") + 1] != NET_PROTOCOL_VERSION)
 			{
-			case CCREQ_CONNECT:
 				if (developer.integer >= 10)
-					Con_Printf("Datagram_ParseConnectionless: received CCREQ_CONNECT from %s.\n", addressstring2);
-				if (length >= (int)strlen("QUAKE") + 1 + 1)
-				{
-					if (memcmp(data, "QUAKE", strlen("QUAKE") + 1) != 0 || (int)data[strlen("QUAKE") + 1] != NET_PROTOCOL_VERSION)
-					{
-						if (developer.integer >= 10)
-							Con_Printf("Datagram_ParseConnectionless: sending CCREP_REJECT \"Incompatible version.\" to %s.\n", addressstring2);
-						SZ_Clear(&net_message);
-						// save space for the header, filled in later
-						MSG_WriteLong(&net_message, 0);
-						MSG_WriteByte(&net_message, CCREP_REJECT);
-						MSG_WriteString(&net_message, "Incompatible version.\n");
-						*((int *)net_message.data) = BigLong(NETFLAG_CTL | (net_message.cursize & NETFLAG_LENGTH_MASK));
-						NetConn_Write(mysocket, net_message.data, net_message.cursize, peeraddress);
-						SZ_Clear(&net_message);
-					}
-					else
-					{
-						// see if this is a duplicate connection request
-						for (clientnum = 0, client = svs.clients;clientnum < svs.maxclients;clientnum++, client++)
-							if (client->netconnection && LHNETADDRESS_Compare(peeraddress, &client->netconnection->peeraddress) == 0)
-								break;
-						if (clientnum < svs.maxclients)
-						{
-							// duplicate connection request
-							if (realtime - client->connecttime < 2.0)
-							{
-								// client is still trying to connect,
-								// so we send a duplicate reply
-								if (developer.integer >= 10)
-									Con_Printf("Datagram_ParseConnectionless: sending duplicate CCREP_ACCEPT to %s.\n", addressstring2);
-								SZ_Clear(&net_message);
-								// save space for the header, filled in later
-								MSG_WriteLong(&net_message, 0);
-								MSG_WriteByte(&net_message, CCREP_ACCEPT);
-								MSG_WriteLong(&net_message, LHNETADDRESS_GetPort(LHNET_AddressFromSocket(client->netconnection->mysocket)));
-								*((int *)net_message.data) = BigLong(NETFLAG_CTL | (net_message.cursize & NETFLAG_LENGTH_MASK));
-								NetConn_Write(mysocket, net_message.data, net_message.cursize, peeraddress);
-								SZ_Clear(&net_message);
-							}
-#if 0
-							else if (realtime - client->netconnection->lastMessageTime >= net_messagerejointimeout.value)
-							{
-								SV_SendServerinfo(client);
-								// the old client hasn't sent us anything
-								// in quite a while, so kick off and let
-								// the retry take care of it...
-								//host_client = client;
-								//SV_DropClient (true);
-							}
-#endif
-						}
-						else
-						{
-							// this is a new client, find a slot
-							for (clientnum = 0, client = svs.clients;clientnum < svs.maxclients;clientnum++, client++)
-								if (!client->active)
-									break;
-							if (clientnum < svs.maxclients && (client->netconnection = conn = NetConn_Open(mysocket, peeraddress)) != NULL)
-							{
-								// connect to the client
-								// everything is allocated, just fill in the details
-								strlcpy (conn->address, addressstring2, sizeof (conn->address));
-								if (developer.integer >= 10)
-									Con_Printf("Datagram_ParseConnectionless: sending CCREP_ACCEPT to %s.\n", addressstring2);
-								// send back the info about the server connection
-								SZ_Clear(&net_message);
-								// save space for the header, filled in later
-								MSG_WriteLong(&net_message, 0);
-								MSG_WriteByte(&net_message, CCREP_ACCEPT);
-								MSG_WriteLong(&net_message, LHNETADDRESS_GetPort(LHNET_AddressFromSocket(conn->mysocket)));
-								*((int *)net_message.data) = BigLong(NETFLAG_CTL | (net_message.cursize & NETFLAG_LENGTH_MASK));
-								NetConn_Write(mysocket, net_message.data, net_message.cursize, peeraddress);
-								SZ_Clear(&net_message);
-								// now set up the client struct
-								SV_VM_Begin();
-								SV_ConnectClient(clientnum, conn);
-								SV_VM_End();
-								NetConn_Heartbeat(1);
-							}
-							else
-							{
-								if (developer.integer >= 10)
-									Con_Printf("Datagram_ParseConnectionless: sending CCREP_REJECT \"Server is full.\" to %s.\n", addressstring2);
-								// no room; try to let player know
-								SZ_Clear(&net_message);
-								// save space for the header, filled in later
-								MSG_WriteLong(&net_message, 0);
-								MSG_WriteByte(&net_message, CCREP_REJECT);
-								MSG_WriteString(&net_message, "Server is full.\n");
-								*((int *)net_message.data) = BigLong(NETFLAG_CTL | (net_message.cursize & NETFLAG_LENGTH_MASK));
-								NetConn_Write(mysocket, net_message.data, net_message.cursize, peeraddress);
-								SZ_Clear(&net_message);
-							}
-						}
-					}
-				}
-				break;
-#if 0
-			case CCREQ_SERVER_INFO:
-				if (developer.integer >= 10)
-					Con_Printf("Datagram_ParseConnectionless: received CCREQ_SERVER_INFO from %s.\n", addressstring2);
-				if (sv.active && !strcmp(MSG_ReadString(), "QUAKE"))
-				{
-					if (developer.integer >= 10)
-						Con_Printf("Datagram_ParseConnectionless: sending CCREP_SERVER_INFO to %s.\n", addressstring2);
-					SZ_Clear(&net_message);
-					// save space for the header, filled in later
-					MSG_WriteLong(&net_message, 0);
-					MSG_WriteByte(&net_message, CCREP_SERVER_INFO);
-					UDP_GetSocketAddr(UDP_acceptSock, &newaddr);
-					MSG_WriteString(&net_message, UDP_AddrToString(&newaddr));
-					MSG_WriteString(&net_message, hostname.string);
-					MSG_WriteString(&net_message, sv.name);
-					MSG_WriteByte(&net_message, net_activeconnections);
-					MSG_WriteByte(&net_message, svs.maxclients);
-					MSG_WriteByte(&net_message, NET_PROTOCOL_VERSION);
-					*((int *)net_message.data) = BigLong(NETFLAG_CTL | (net_message.cursize & NETFLAG_LENGTH_MASK));
-					NetConn_Write(mysocket, net_message.data, net_message.cursize, peeraddress);
-					SZ_Clear(&net_message);
-				}
-				break;
-			case CCREQ_PLAYER_INFO:
-				if (developer.integer >= 10)
-					Con_Printf("Datagram_ParseConnectionless: received CCREQ_PLAYER_INFO from %s.\n", addressstring2);
-				if (sv.active)
-				{
-					int playerNumber, activeNumber, clientNumber;
-					client_t *client;
-
-					playerNumber = MSG_ReadByte();
-					activeNumber = -1;
-					for (clientNumber = 0, client = svs.clients; clientNumber < svs.maxclients; clientNumber++, client++)
-						if (client->active && ++activeNumber == playerNumber)
-							break;
-					if (clientNumber != svs.maxclients)
-					{
-						SZ_Clear(&net_message);
-						// save space for the header, filled in later
-						MSG_WriteLong(&net_message, 0);
-						MSG_WriteByte(&net_message, CCREP_PLAYER_INFO);
-						MSG_WriteByte(&net_message, playerNumber);
-						MSG_WriteString(&net_message, client->name);
-						MSG_WriteLong(&net_message, client->colors);
-						MSG_WriteLong(&net_message, (int)client->edict->fields.server->frags);
-						MSG_WriteLong(&net_message, (int)(realtime - client->connecttime));
-						MSG_WriteString(&net_message, client->netconnection ? client->netconnection->address : "botclient");
-						*((int *)net_message.data) = BigLong(NETFLAG_CTL | (net_message.cursize & NETFLAG_LENGTH_MASK));
-						NetConn_Write(mysocket, net_message.data, net_message.cursize, peeraddress);
-						SZ_Clear(&net_message);
-					}
-				}
-				break;
-			case CCREQ_RULE_INFO:
-				if (developer.integer >= 10)
-					Con_Printf("Datagram_ParseConnectionless: received CCREQ_RULE_INFO from %s.\n", addressstring2);
-				if (sv.active)
-				{
-					char *prevCvarName;
-					cvar_t *var;
-
-					// find the search start location
-					prevCvarName = MSG_ReadString();
-					var = Cvar_FindVarAfter(prevCvarName, CVAR_NOTIFY);
-
-					// send the response
-					SZ_Clear(&net_message);
-					// save space for the header, filled in later
-					MSG_WriteLong(&net_message, 0);
-					MSG_WriteByte(&net_message, CCREP_RULE_INFO);
-					if (var)
-					{
-						MSG_WriteString(&net_message, var->name);
-						MSG_WriteString(&net_message, var->string);
-					}
-					*((int *)net_message.data) = BigLong(NETFLAG_CTL | (net_message.cursize & NETFLAG_LENGTH_MASK));
-					NetConn_Write(mysocket, net_message.data, net_message.cursize, peeraddress);
-					SZ_Clear(&net_message);
-				}
-				break;
-#endif
-			default:
+					Con_Printf("Datagram_ParseConnectionless: sending CCREP_REJECT \"Incompatible version.\" to %s.\n", addressstring2);
+				SZ_Clear(&net_message);
+				// save space for the header, filled in later
+				MSG_WriteLong(&net_message, 0);
+				MSG_WriteByte(&net_message, CCREP_REJECT);
+				MSG_WriteString(&net_message, "Incompatible version.\n");
+				*((int *)net_message.data) = BigLong(NETFLAG_CTL | (net_message.cursize & NETFLAG_LENGTH_MASK));
+				NetConn_Write(mysocket, net_message.data, net_message.cursize, peeraddress);
+				SZ_Clear(&net_message);
 				break;
 			}
-			// we may not have liked the packet, but it was a valid control
-			// packet, so we're done processing this packet now
-			return true;
+
+			// see if this connect request comes from a known client
+			for (clientnum = 0, client = svs.clients;clientnum < svs.maxclients;clientnum++, client++)
+			{
+				if (client->netconnection && LHNETADDRESS_Compare(peeraddress, &client->netconnection->peeraddress) == 0)
+				{
+					// this is either a duplicate connection request
+					// or coming back from a timeout
+					// (if so, keep their stuff intact)
+
+					// send a reply
+					if (developer.integer >= 10)
+						Con_Printf("Datagram_ParseConnectionless: sending duplicate CCREP_ACCEPT to %s.\n", addressstring2);
+					SZ_Clear(&net_message);
+					// save space for the header, filled in later
+					MSG_WriteLong(&net_message, 0);
+					MSG_WriteByte(&net_message, CCREP_ACCEPT);
+					MSG_WriteLong(&net_message, LHNETADDRESS_GetPort(LHNET_AddressFromSocket(client->netconnection->mysocket)));
+					*((int *)net_message.data) = BigLong(NETFLAG_CTL | (net_message.cursize & NETFLAG_LENGTH_MASK));
+					NetConn_Write(mysocket, net_message.data, net_message.cursize, peeraddress);
+					SZ_Clear(&net_message);
+
+					// if client is already spawned, re-send the
+					// serverinfo message as they'll need it to play
+					if (client->spawned)
+					{
+						SV_VM_Begin();
+						SV_SendServerinfo(client);
+						SV_VM_End();
+					}
+					return true;
+				}
+			}
+
+			// this is a new client, check for connection flood
+			if (NetConn_PreventConnectFlood(peeraddress))
+				break;
+
+			// find a slot for the new client
+			for (clientnum = 0, client = svs.clients;clientnum < svs.maxclients;clientnum++, client++)
+			{
+				netconn_t *conn;
+				if (!client->active && (client->netconnection = conn = NetConn_Open(mysocket, peeraddress)) != NULL)
+				{
+					// connect to the client
+					// everything is allocated, just fill in the details
+					strlcpy (conn->address, addressstring2, sizeof (conn->address));
+					if (developer.integer >= 10)
+						Con_Printf("Datagram_ParseConnectionless: sending CCREP_ACCEPT to %s.\n", addressstring2);
+					// send back the info about the server connection
+					SZ_Clear(&net_message);
+					// save space for the header, filled in later
+					MSG_WriteLong(&net_message, 0);
+					MSG_WriteByte(&net_message, CCREP_ACCEPT);
+					MSG_WriteLong(&net_message, LHNETADDRESS_GetPort(LHNET_AddressFromSocket(conn->mysocket)));
+					*((int *)net_message.data) = BigLong(NETFLAG_CTL | (net_message.cursize & NETFLAG_LENGTH_MASK));
+					NetConn_Write(mysocket, net_message.data, net_message.cursize, peeraddress);
+					SZ_Clear(&net_message);
+					// now set up the client struct
+					SV_VM_Begin();
+					SV_ConnectClient(clientnum, conn);
+					SV_VM_End();
+					NetConn_Heartbeat(1);
+					return true;
+				}
+			}
+
+			if (developer.integer >= 10)
+				Con_Printf("Datagram_ParseConnectionless: sending CCREP_REJECT \"Server is full.\" to %s.\n", addressstring2);
+			// no room; try to let player know
+			SZ_Clear(&net_message);
+			// save space for the header, filled in later
+			MSG_WriteLong(&net_message, 0);
+			MSG_WriteByte(&net_message, CCREP_REJECT);
+			MSG_WriteString(&net_message, "Server is full.\n");
+			*((int *)net_message.data) = BigLong(NETFLAG_CTL | (net_message.cursize & NETFLAG_LENGTH_MASK));
+			NetConn_Write(mysocket, net_message.data, net_message.cursize, peeraddress);
+			SZ_Clear(&net_message);
+			break;
+		case CCREQ_SERVER_INFO:
+			if (developer.integer >= 10)
+				Con_Printf("Datagram_ParseConnectionless: received CCREQ_SERVER_INFO from %s.\n", addressstring2);
+			if (sv.active && !strcmp(MSG_ReadString(), "QUAKE"))
+			{
+				int numclients;
+				char myaddressstring[128];
+				if (developer.integer >= 10)
+					Con_Printf("Datagram_ParseConnectionless: sending CCREP_SERVER_INFO to %s.\n", addressstring2);
+				SZ_Clear(&net_message);
+				// save space for the header, filled in later
+				MSG_WriteLong(&net_message, 0);
+				MSG_WriteByte(&net_message, CCREP_SERVER_INFO);
+				LHNETADDRESS_ToString(LHNET_AddressFromSocket(mysocket), myaddressstring, sizeof(myaddressstring), true);
+				MSG_WriteString(&net_message, myaddressstring);
+				MSG_WriteString(&net_message, hostname.string);
+				MSG_WriteString(&net_message, sv.name);
+				// How many clients are there?
+				for (i = 0, numclients = 0;i < svs.maxclients;i++)
+					if (svs.clients[i].active)
+						numclients++;
+				MSG_WriteByte(&net_message, numclients);
+				MSG_WriteByte(&net_message, svs.maxclients);
+				MSG_WriteByte(&net_message, NET_PROTOCOL_VERSION);
+				*((int *)net_message.data) = BigLong(NETFLAG_CTL | (net_message.cursize & NETFLAG_LENGTH_MASK));
+				NetConn_Write(mysocket, net_message.data, net_message.cursize, peeraddress);
+				SZ_Clear(&net_message);
+			}
+			break;
+		case CCREQ_PLAYER_INFO:
+			if (developer.integer >= 10)
+				Con_Printf("Datagram_ParseConnectionless: received CCREQ_PLAYER_INFO from %s.\n", addressstring2);
+			if (sv.active)
+			{
+				int playerNumber, activeNumber, clientNumber;
+				client_t *client;
+
+				playerNumber = MSG_ReadByte();
+				activeNumber = -1;
+				for (clientNumber = 0, client = svs.clients; clientNumber < svs.maxclients; clientNumber++, client++)
+					if (client->active && ++activeNumber == playerNumber)
+						break;
+				if (clientNumber != svs.maxclients)
+				{
+					SZ_Clear(&net_message);
+					// save space for the header, filled in later
+					MSG_WriteLong(&net_message, 0);
+					MSG_WriteByte(&net_message, CCREP_PLAYER_INFO);
+					MSG_WriteByte(&net_message, playerNumber);
+					MSG_WriteString(&net_message, client->name);
+					MSG_WriteLong(&net_message, client->colors);
+					MSG_WriteLong(&net_message, (int)client->edict->fields.server->frags);
+					MSG_WriteLong(&net_message, (int)(realtime - client->connecttime));
+					MSG_WriteString(&net_message, client->netconnection ? client->netconnection->address : "botclient");
+					*((int *)net_message.data) = BigLong(NETFLAG_CTL | (net_message.cursize & NETFLAG_LENGTH_MASK));
+					NetConn_Write(mysocket, net_message.data, net_message.cursize, peeraddress);
+					SZ_Clear(&net_message);
+				}
+			}
+			break;
+		case CCREQ_RULE_INFO:
+			if (developer.integer >= 10)
+				Con_Printf("Datagram_ParseConnectionless: received CCREQ_RULE_INFO from %s.\n", addressstring2);
+			if (sv.active)
+			{
+				char *prevCvarName;
+				cvar_t *var;
+
+				// find the search start location
+				prevCvarName = MSG_ReadString();
+				var = Cvar_FindVarAfter(prevCvarName, CVAR_NOTIFY);
+
+				// send the response
+				SZ_Clear(&net_message);
+				// save space for the header, filled in later
+				MSG_WriteLong(&net_message, 0);
+				MSG_WriteByte(&net_message, CCREP_RULE_INFO);
+				if (var)
+				{
+					MSG_WriteString(&net_message, var->name);
+					MSG_WriteString(&net_message, var->string);
+				}
+				*((int *)net_message.data) = BigLong(NETFLAG_CTL | (net_message.cursize & NETFLAG_LENGTH_MASK));
+				NetConn_Write(mysocket, net_message.data, net_message.cursize, peeraddress);
+				SZ_Clear(&net_message);
+			}
+			break;
+		default:
+			break;
 		}
+		// we may not have liked the packet, but it was a valid control
+		// packet, so we're done processing this packet now
+		return true;
 	}
-#endif
 	if (host_client)
 	{
-		if ((ret = NetConn_ReceivedMessage(host_client->netconnection, data, length, sv.protocol)) == 2)
+		if ((ret = NetConn_ReceivedMessage(host_client->netconnection, data, length, sv.protocol, host_client->spawned ? net_messagetimeout.value : net_connecttimeout.value)) == 2)
 		{
 			SV_VM_Begin();
 			SV_ReadClientMessage();
@@ -2364,8 +2463,8 @@ void NetConn_Init(void)
 	Cvar_RegisterVariable(&net_slist_timeout);
 	Cvar_RegisterVariable(&net_slist_maxtries);
 	Cvar_RegisterVariable(&net_messagetimeout);
-	Cvar_RegisterVariable(&net_messagerejointimeout);
 	Cvar_RegisterVariable(&net_connecttimeout);
+	Cvar_RegisterVariable(&net_connectfloodblockingtimeout);
 	Cvar_RegisterVariable(&cl_netlocalping);
 	Cvar_RegisterVariable(&cl_netpacketloss);
 	Cvar_RegisterVariable(&hostname);
