@@ -26,7 +26,7 @@ Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 
 
 #define SND_MIN_SPEED 8000
-#define SND_MAX_SPEED 96000
+#define SND_MAX_SPEED 48000
 #define SND_MIN_WIDTH 1
 #define SND_MAX_WIDTH 2
 #define SND_MIN_CHANNELS 1
@@ -146,7 +146,6 @@ static double snd_starttime = 0.0;
 
 vec3_t listener_origin;
 matrix4x4_t listener_matrix[SND_LISTENERS];
-vec_t sound_nominal_clip_dist=1000.0;
 mempool_t *snd_mempool;
 
 // Linked list of known sfx
@@ -168,6 +167,7 @@ cvar_t bgmvolume = {CVAR_SAVE, "bgmvolume", "1", "volume of background music (su
 cvar_t volume = {CVAR_SAVE, "volume", "0.7", "volume of sound effects"};
 cvar_t snd_initialized = { CVAR_READONLY, "snd_initialized", "0", "indicates the sound subsystem is active"};
 cvar_t snd_staticvolume = {CVAR_SAVE, "snd_staticvolume", "1", "volume of ambient sound effects (such as swampy sounds at the start of e1m2)"};
+cvar_t snd_soundradius = {0, "snd_soundradius", "1000", "radius of weapon sounds and other standard sound effects (monster idle noises are half this radius and flickering light noises are one third of this radius)"};
 
 // Cvars declared in snd_main.h (shared with other snd_*.c files)
 cvar_t _snd_mixahead = {CVAR_SAVE, "_snd_mixahead", "0.1", "how much sound to mix ahead of time"};
@@ -301,54 +301,79 @@ void S_SoundInfo_f(void)
 }
 
 
-// TODO: make this function smarter...
+int S_GetSoundRate(void)
+{
+	return snd_renderbuffer ? snd_renderbuffer->format.speed : 0;
+}
+
+
 static qboolean S_ChooseCheaperFormat (snd_format_t* format, qboolean fixed_speed, qboolean fixed_width, qboolean fixed_channels)
 {
-	// Can we decrease the number of channels?
-	if (!fixed_channels && format->channels > 1)
+	static const snd_format_t thresholds [] =
 	{
-		unsigned short channels = format->channels;
+		// speed			width			channels
+		{ SND_MIN_SPEED,	SND_MIN_WIDTH,	SND_MIN_CHANNELS },
+		{ 11025,			1,				2 },
+		{ 22050,			2,				2 },
+		{ 44100,			2,				2 },
+		{ 48000,			2,				6 },
+		{ SND_MAX_SPEED,	SND_MAX_WIDTH,	SND_MAX_CHANNELS },
+	};
+	const unsigned int nb_thresholds = sizeof(thresholds) / sizeof(thresholds[0]);
+	unsigned int speed_level, width_level, channels_level;
 
-		// If it has an odd number of channels(?!), make it even
-		if (channels & 1)
-			channels--;
-		else
-		{
-			// Remove 2 speakers, unless it's a stereo format
-			if (channels != 2)
-				channels -= 2;
-			else
-				channels = 1;
-		}
+	// If we have reached the minimum values, there's nothing more we can do
+	if ((format->speed == thresholds[0].speed || fixed_speed) &&
+		(format->width == thresholds[0].width || fixed_width) &&
+		(format->channels == thresholds[0].channels || fixed_channels))
+		return false;
 
-		format->channels = channels;
+	// Check the min and max values
+	#define CHECK_BOUNDARIES(param)								\
+	if (format->param < thresholds[0].param)					\
+	{															\
+		format->param = thresholds[0].param;					\
+		return true;											\
+	}															\
+	if (format->param > thresholds[nb_thresholds - 1].param)	\
+	{															\
+		format->param = thresholds[nb_thresholds - 1].param;	\
+		return true;											\
+	}
+	CHECK_BOUNDARIES(speed);
+	CHECK_BOUNDARIES(width);
+	CHECK_BOUNDARIES(channels);
+	#undef CHECK_BOUNDARIES
+
+	// Find the level of each parameter
+	#define FIND_LEVEL(param)									\
+	param##_level = 0;											\
+	while (param##_level < nb_thresholds - 1)					\
+	{															\
+		if (format->param <= thresholds[param##_level].param)	\
+			break;												\
+																\
+		param##_level++;										\
+	}
+	FIND_LEVEL(speed);
+	FIND_LEVEL(width);
+	FIND_LEVEL(channels);
+	#undef FIND_LEVEL
+
+	// Decrease the parameter with the highest level to the previous level
+	if (channels_level >= speed_level && channels_level >= width_level && !fixed_channels)
+	{
+		format->channels = thresholds[channels_level - 1].channels;
+		return true;
+	}
+	if (speed_level >= width_level && !fixed_speed)
+	{
+		format->speed = thresholds[speed_level - 1].speed;
 		return true;
 	}
 
-	// Can we decrease the speed?
-	if (!fixed_speed)
-	{
-		unsigned int suggest_speeds [] = { 44100, 22050, 11025 };
-		unsigned int i;
-
-		for (i = 0; i < sizeof(suggest_speeds) / sizeof(suggest_speeds[0]); i++)
-			if (format->speed > suggest_speeds[i])
-			{
-				format->speed = suggest_speeds[i];
-				return true;
-			}
-
-		// the speed is already low
-	}
-
-	// Can we decrease the number of bits per sample?
-	if (!fixed_width && format->width > 1)
-	{
-		format->width = 1;
-		return true;
-	}
-
-	return false;
+	format->width = thresholds[width_level - 1].width;
+	return true;
 }
 
 
@@ -401,7 +426,7 @@ static void S_SetChannelLayout (void)
 	if (snd_channellayout.integer < SND_CHANNELLAYOUT_AUTO ||
 		snd_channellayout.integer > SND_CHANNELLAYOUT_ALSA)
 		Cvar_SetValueQuick (&snd_channellayout, SND_CHANNELLAYOUT_STANDARD);
-	
+
 	if (snd_channellayout.integer == SND_CHANNELLAYOUT_AUTO)
 	{
 		// If we're in the sound engine initialization
@@ -608,9 +633,12 @@ void S_Startup (void)
 			   chosen_fmt.speed, chosen_fmt.channels, chosen_fmt.width * 8);
 
 	// Update the cvars
-	snd_speed.integer = chosen_fmt.speed;
-	snd_width.integer = chosen_fmt.width;
-	snd_channels.integer = chosen_fmt.channels;
+	if (snd_speed.integer != (int)chosen_fmt.speed)
+		Cvar_SetValueQuick(&snd_speed, chosen_fmt.speed);
+	if (snd_width.integer != chosen_fmt.width)
+		Cvar_SetValueQuick(&snd_width, chosen_fmt.width);
+	if (snd_channels.integer != chosen_fmt.channels)
+		Cvar_SetValueQuick(&snd_channels, chosen_fmt.channels);
 
 	current_channellayout_used = SND_CHANNELLAYOUT_AUTO;
 	S_SetChannelLayout();
@@ -708,6 +736,7 @@ void S_Init(void)
 	Cvar_RegisterVariable(&_snd_mixahead);
 	Cvar_RegisterVariable(&snd_swapstereo); // for people with backwards sound wiring
 	Cvar_RegisterVariable(&snd_channellayout);
+	Cvar_RegisterVariable(&snd_soundradius);
 
 	Cvar_SetValueQuick(&snd_initialized, true);
 
@@ -1112,10 +1141,10 @@ void S_PlaySfxOnChannel (sfx_t *sfx, channel_t *target_chan, unsigned int flags,
 	{
 		if (sfx->loopstart == -1)
 			Con_DPrintf("Quake compatibility warning: Static sound \"%s\" is not looped\n", sfx->name);
-		target_chan->dist_mult = attenuation / (64.0f * sound_nominal_clip_dist);
+		target_chan->dist_mult = attenuation / (64.0f * snd_soundradius.value);
 	}
 	else
-		target_chan->dist_mult = attenuation / sound_nominal_clip_dist;
+		target_chan->dist_mult = attenuation / snd_soundradius.value;
 
 	// Lock the SFX during play
 	S_LockSfx (sfx);
@@ -1362,13 +1391,13 @@ static void S_PaintAndSubmit (void)
 
 	if (snd_renderbuffer == NULL || nosound.integer)
 		return;
-	
-	if (snd_blocked > 0 && !cls.capturevideo_soundfile)
+
+	if (snd_blocked > 0 && !(cls.capturevideo.soundrate && !cls.capturevideo.realtime))
 		return;
 
 	// Update sound time
-	if (cls.capturevideo_soundfile) // SUPER NASTY HACK to record non-realtime sound
-		newsoundtime = (unsigned int)((double)cls.capturevideo_frame * (double)snd_renderbuffer->format.speed / (double)cls.capturevideo_framerate);
+	if (cls.capturevideo.soundrate && !cls.capturevideo.realtime) // SUPER NASTY HACK to record non-realtime sound
+		newsoundtime = (unsigned int)((double)cls.capturevideo.frame * (double)snd_renderbuffer->format.speed / (double)cls.capturevideo.framerate);
 	else if (simsound)
 		newsoundtime = (unsigned int)((realtime - snd_starttime) * (double)snd_renderbuffer->format.speed);
 	else
@@ -1377,7 +1406,7 @@ static void S_PaintAndSubmit (void)
 	newsoundtime += extrasoundtime;
 	if (newsoundtime < soundtime)
 	{
-		if ((cls.capturevideo_soundfile != NULL) != recording_sound)
+		if ((cls.capturevideo.soundrate != 0) != recording_sound)
 		{
 			unsigned int additionaltime;
 
@@ -1388,7 +1417,7 @@ static void S_PaintAndSubmit (void)
 			// some modules write directly to a shared (DMA) buffer
 			additionaltime = (soundtime - newsoundtime) + snd_renderbuffer->maxframes - 1;
 			additionaltime -= additionaltime % snd_renderbuffer->maxframes;
-			
+
 			extrasoundtime += additionaltime;
 			newsoundtime += additionaltime;
 			Con_DPrintf("S_PaintAndSubmit: new extra sound time = %u\n",
@@ -1399,7 +1428,7 @@ static void S_PaintAndSubmit (void)
 					   newsoundtime, soundtime);
 	}
 	soundtime = newsoundtime;
-	recording_sound = (cls.capturevideo_soundfile != NULL);
+	recording_sound = (cls.capturevideo.soundrate != 0);
 
 	// Check to make sure that we haven't overshot
 	paintedtime = snd_renderbuffer->endframe;
@@ -1435,8 +1464,8 @@ void S_Update(const matrix4x4_t *listenermatrix)
 
 	if (snd_renderbuffer == NULL || nosound.integer)
 		return;
-	
-	if (snd_blocked > 0 && !cls.capturevideo_soundfile)
+
+	if (snd_blocked > 0 && !(cls.capturevideo.soundrate && !cls.capturevideo.realtime))
 		return;
 
 	// If snd_swapstereo or snd_channellayout has changed, recompute the channel layout
