@@ -71,13 +71,14 @@ cvar_t vid_width = {CVAR_SAVE, "vid_width", "640", "resolution"};
 cvar_t vid_height = {CVAR_SAVE, "vid_height", "480", "resolution"};
 cvar_t vid_bitsperpixel = {CVAR_SAVE, "vid_bitsperpixel", "32", "how many bits per pixel to render at (32 or 16, 32 is recommended)"};
 cvar_t vid_refreshrate = {CVAR_SAVE, "vid_refreshrate", "60", "refresh rate to use, in hz (higher values flicker less, if supported by your monitor)"};
+cvar_t vid_stereobuffer = {CVAR_SAVE, "vid_stereobuffer", "0", "enables 'quad-buffered' stereo rendering for stereo shutterglasses, HMD (head mounted display) devices, or polarized stereo LCDs, if supported by your drivers"};
 
 cvar_t vid_vsync = {CVAR_SAVE, "vid_vsync", "0", "sync to vertical blank, prevents 'tearing' (seeing part of one frame and part of another on the screen at the same time), automatically disabled when doing timedemo benchmarks"};
 cvar_t vid_mouse = {CVAR_SAVE, "vid_mouse", "1", "whether to use the mouse in windowed mode (fullscreen always does)"};
 cvar_t vid_minwidth = {0, "vid_minwidth", "0", "minimum vid_width that is acceptable (to be set in default.cfg in mods)"};
 cvar_t vid_minheight = {0, "vid_minheight", "0", "minimum vid_height that is acceptable (to be set in default.cfg in mods)"};
 cvar_t gl_combine = {0, "gl_combine", "1", "faster rendering by using GL_ARB_texture_env_combine extension (part of OpenGL 1.3 and above)"};
-cvar_t gl_finish = {0, "gl_finish", "0", "make the cpu wait for the graphics processor at the end of each frame render (can help with strange input or video lag problems on some machines)"};
+cvar_t gl_finish = {0, "gl_finish", "0", "make the cpu wait for the graphics processor at the end of each rendered frame (can help with strange input or video lag problems on some machines)"};
 
 cvar_t v_gamma = {CVAR_SAVE, "v_gamma", "1", "inverse gamma correction value, a brightness effect that does not affect white or black, and tends to make the image grey and dull"};
 cvar_t v_contrast = {CVAR_SAVE, "v_contrast", "1", "brightness of white (values above 1 give a brighter image with increased color saturation, unlike v_gamma)"};
@@ -92,7 +93,7 @@ cvar_t v_color_grey_b = {CVAR_SAVE, "v_color_grey_b", "0.5", "desired color of g
 cvar_t v_color_white_r = {CVAR_SAVE, "v_color_white_r", "1", "desired color of white"};
 cvar_t v_color_white_g = {CVAR_SAVE, "v_color_white_g", "1", "desired color of white"};
 cvar_t v_color_white_b = {CVAR_SAVE, "v_color_white_b", "1", "desired color of white"};
-cvar_t v_hwgamma = {CVAR_SAVE, "v_hwgamma", "1", "enables use of hardware gamma correction ramps if available (note: does not work very well on Windows2000 and above)"};
+cvar_t v_hwgamma = {CVAR_SAVE, "v_hwgamma", "1", "enables use of hardware gamma correction ramps if available (note: does not work very well on Windows2000 and above), values are 0 = off, 1 = attempt to use hardware gamma, 2 = use hardware gamma whether it works or not"};
 cvar_t v_psycho = {0, "v_psycho", "0", "easter egg (does not work on Windows2000 or above)"};
 
 // brand of graphics chip
@@ -239,6 +240,7 @@ void (GLAPIENTRY *qglScissor)(GLint x, GLint y, GLsizei width, GLsizei height);
 
 void (GLAPIENTRY *qglPolygonOffset)(GLfloat factor, GLfloat units);
 void (GLAPIENTRY *qglPolygonMode)(GLenum face, GLenum mode);
+void (GLAPIENTRY *qglPolygonStipple)(const GLubyte *mask);
 
 //[515]: added on 29.07.2005
 void (GLAPIENTRY *qglLineWidth)(GLfloat width);
@@ -467,6 +469,7 @@ static dllfunction_t opengl110funcs[] =
 	{"glScissor", (void **) &qglScissor},
 	{"glPolygonOffset", (void **) &qglPolygonOffset},
 	{"glPolygonMode", (void **) &qglPolygonMode},
+	{"glPolygonStipple", (void **) &qglPolygonStipple},
 	{NULL, NULL}
 };
 
@@ -716,7 +719,6 @@ void VID_UpdateGamma(qboolean force, int rampsize)
 {
 	cvar_t *c;
 	float f;
-	static int forcenextframe = false;
 
 	// LordHavoc: don't mess with gamma tables if running dedicated
 	if (cls.state == ca_dedicated)
@@ -725,7 +727,7 @@ void VID_UpdateGamma(qboolean force, int rampsize)
 	if (!force
 	 && !gamma_forcenextframe
 	 && !v_psycho.integer
-	 && cachehwgamma == (vid_activewindow && v_hwgamma.integer)
+	 && cachehwgamma == (vid_activewindow ? v_hwgamma.integer : 0)
 	 && v_gamma.value == cachegamma
 	 && v_contrast.value == cachecontrast
 	 && v_brightness.value == cachebrightness
@@ -754,9 +756,9 @@ void VID_UpdateGamma(qboolean force, int rampsize)
 	BOUNDCVAR(v_color_white_g, 1, 5);cachewhite[1] = v_color_white_g.value;
 	BOUNDCVAR(v_color_white_b, 1, 5);cachewhite[2] = v_color_white_b.value;
 	cachecolorenable = v_color_enable.integer;
-	cachehwgamma = vid_activewindow && v_hwgamma.integer;
+	cachehwgamma = vid_activewindow ? v_hwgamma.integer : 0;
 
-	forcenextframe = false;
+	gamma_forcenextframe = false;
 
 	if (cachehwgamma)
 	{
@@ -797,7 +799,7 @@ void VID_UpdateGamma(qboolean force, int rampsize)
 			static float n[3], nd[3], nt[3];
 			static int init = true;
 			unsigned short *ramp;
-			forcenextframe = true;
+			gamma_forcenextframe = true;
 			if (init)
 			{
 				init = false;
@@ -826,16 +828,26 @@ void VID_UpdateGamma(qboolean force, int rampsize)
 					*ramp++ = (unsigned short)(cos(t*(M_PI*2.0)) * 32767.0f + 32767.0f);
 		}
 
-		Cvar_SetValueQuick(&vid_hardwaregammasupported, VID_SetGamma(vid_gammaramps, vid_gammarampsize));
+		// set vid_hardwaregammasupported to true if VID_SetGamma succeeds, OR if vid_hwgamma is >= 2 (forced gamma - ignores driver return value)
+		Cvar_SetValueQuick(&vid_hardwaregammasupported, VID_SetGamma(vid_gammaramps, vid_gammarampsize) || cachehwgamma >= 2);
 		// if custom gamma ramps failed (Windows stupidity), restore to system gamma
 		if(!vid_hardwaregammasupported.integer)
 		{
-			VID_RestoreSystemGamma();
-			Cvar_SetValueQuick(&vid_hardwaregammasupported, false);
+			if (vid_usinghwgamma)
+			{
+				vid_usinghwgamma = false;
+				VID_SetGamma(vid_systemgammaramps, vid_gammarampsize);
+			}
 		}
 	}
 	else
-		VID_RestoreSystemGamma();
+	{
+		if (vid_usinghwgamma)
+		{
+			vid_usinghwgamma = false;
+			VID_SetGamma(vid_systemgammaramps, vid_gammarampsize);
+		}
+	}
 }
 
 void VID_RestoreSystemGamma(void)
@@ -876,6 +888,7 @@ void VID_Shared_Init(void)
 	Cvar_RegisterVariable(&vid_height);
 	Cvar_RegisterVariable(&vid_bitsperpixel);
 	Cvar_RegisterVariable(&vid_refreshrate);
+	Cvar_RegisterVariable(&vid_stereobuffer);
 	Cvar_RegisterVariable(&vid_vsync);
 	Cvar_RegisterVariable(&vid_mouse);
 	Cvar_RegisterVariable(&vid_minwidth);
@@ -888,21 +901,23 @@ void VID_Shared_Init(void)
 		Cvar_Set("gl_combine", "0");
 }
 
-int VID_Mode(int fullscreen, int width, int height, int bpp, int refreshrate)
+int VID_Mode(int fullscreen, int width, int height, int bpp, int refreshrate, int stereobuffer)
 {
-	Con_Printf("Video: %s %dx%dx%dx%dhz\n", fullscreen ? "fullscreen" : "window", width, height, bpp, refreshrate);
-	if (VID_InitMode(fullscreen, width, height, bpp, refreshrate))
+	Con_Printf("Video: %s %dx%dx%dx%dhz%s\n", fullscreen ? "fullscreen" : "window", width, height, bpp, refreshrate, stereobuffer ? " stereo" : "");
+	if (VID_InitMode(fullscreen, width, height, bpp, refreshrate, stereobuffer))
 	{
 		vid.fullscreen = fullscreen;
 		vid.width = width;
 		vid.height = height;
 		vid.bitsperpixel = bpp;
 		vid.refreshrate = refreshrate;
+		vid.stereobuffer = stereobuffer;
 		Cvar_SetValueQuick(&vid_fullscreen, fullscreen);
 		Cvar_SetValueQuick(&vid_width, width);
 		Cvar_SetValueQuick(&vid_height, height);
 		Cvar_SetValueQuick(&vid_bitsperpixel, bpp);
 		Cvar_SetValueQuick(&vid_refreshrate, refreshrate);
+		Cvar_SetValueQuick(&vid_stereobuffer, stereobuffer);
 		return true;
 	}
 	else
@@ -934,10 +949,10 @@ void VID_Restart_f(void)
 		vid_fullscreen.integer ? "fullscreen" : "window", vid_width.integer, vid_height.integer, vid_bitsperpixel.integer);
 	VID_CloseSystems();
 	VID_Shutdown();
-	if (!VID_Mode(vid_fullscreen.integer, vid_width.integer, vid_height.integer, vid_bitsperpixel.integer, vid_refreshrate.integer))
+	if (!VID_Mode(vid_fullscreen.integer, vid_width.integer, vid_height.integer, vid_bitsperpixel.integer, vid_refreshrate.integer, vid_stereobuffer.integer))
 	{
 		Con_Print("Video mode change failed\n");
-		if (!VID_Mode(vid.fullscreen, vid.width, vid.height, vid.bitsperpixel, vid.refreshrate))
+		if (!VID_Mode(vid.fullscreen, vid.width, vid.height, vid.bitsperpixel, vid.refreshrate, vid.stereobuffer))
 			Sys_Error("Unable to restore to last working video mode");
 	}
 	VID_OpenSystems();
@@ -979,17 +994,19 @@ void VID_Start(void)
 	}
 
 	Con_Print("Starting video system\n");
-	success = VID_Mode(vid_fullscreen.integer, vid_width.integer, vid_height.integer, vid_bitsperpixel.integer, vid_refreshrate.integer);
+	success = VID_Mode(vid_fullscreen.integer, vid_width.integer, vid_height.integer, vid_bitsperpixel.integer, vid_refreshrate.integer, vid_stereobuffer.integer);
 	if (!success)
 	{
 		Con_Print("Desired video mode fail, trying fallbacks...\n");
-		success = VID_Mode(vid_fullscreen.integer, vid_width.integer, vid_height.integer, vid_bitsperpixel.integer, 60);
+		success = VID_Mode(vid_fullscreen.integer, vid_width.integer, vid_height.integer, vid_bitsperpixel.integer, 60, vid_stereobuffer.integer);
+		if (!success && vid_stereobuffer.integer)
+			success = VID_Mode(vid_fullscreen.integer, vid_width.integer, vid_height.integer, vid_bitsperpixel.integer, vid_refreshrate.integer, false);
 		if (!success && vid_bitsperpixel.integer > 16)
-			success = VID_Mode(vid_fullscreen.integer, vid_width.integer, vid_height.integer, 16, 60);
+			success = VID_Mode(vid_fullscreen.integer, vid_width.integer, vid_height.integer, 16, 60, false);
 		if (!success && (vid_width.integer > 640 || vid_height.integer > 480))
-			success = VID_Mode(vid_fullscreen.integer, 640, 480, 16, 60);
+			success = VID_Mode(vid_fullscreen.integer, 640, 480, 16, 60, false);
 		if (!success && vid_fullscreen.integer)
-			success = VID_Mode(false, 640, 480, 16, 60);
+			success = VID_Mode(false, 640, 480, 16, 60, false);
 		if (!success)
 			Sys_Error("Video modes failed");
 	}
