@@ -688,7 +688,7 @@ static int GL_OpenLibrary(const char *name)
 		Con_Printf("Unable to LoadLibrary %s\n", name);
 		return false;
 	}
-	strcpy(gl_driver, name);
+	strlcpy(gl_driver, name, sizeof(gl_driver));
 	return true;
 }
 
@@ -728,7 +728,7 @@ void VID_Init(void)
 	IN_Init();
 }
 
-int VID_InitMode (int fullscreen, int width, int height, int bpp, int refreshrate)
+int VID_InitMode (int fullscreen, int width, int height, int bpp, int refreshrate, int stereobuffer)
 {
 	int i;
 	HDC hdc;
@@ -776,6 +776,9 @@ int VID_InitMode (int fullscreen, int width, int height, int bpp, int refreshrat
 		pfd.cAlphaBits = 0;
 	}
 
+	if (stereobuffer)
+		pfd.dwFlags |= PFD_STEREO;
+
 	gldrivername = "opengl32.dll";
 // COMMANDLINEOPTION: Windows WGL: -gl_driver <drivername> selects a GL driver library, default is opengl32.dll, useful only for 3dfxogl.dll or 3dfxvgl.dll, if you don't know what this is for, you don't need it
 	i = COM_CheckParm("-gl_driver");
