@@ -43,6 +43,8 @@ cvar_t cl_bobmodel_side = {0, "cl_bobmodel_side", "0.05", "gun bobbing sideways
 cvar_t cl_bobmodel_up = {0, "cl_bobmodel_up", "0.02", "gun bobbing upward movement amount"};
 cvar_t cl_bobmodel_speed = {0, "cl_bobmodel_speed", "7", "gun bobbing speed"};
 
+cvar_t cl_viewmodel_scale = {0, "cl_viewmodel_scale", "0.3", "changes size of gun model, lower values prevent poking into walls but cause strange artifacts on lighting and especially r_stereo/vid_stereobuffer options where the size of the gun becomes visible"};
+
 cvar_t v_kicktime = {0, "v_kicktime", "0.5", "how long a view kick from damage lasts"};
 cvar_t v_kickroll = {0, "v_kickroll", "0.6", "how much a view kick from damage rolls your view"};
 cvar_t v_kickpitch = {0, "v_kickpitch", "0.6", "how much a view kick from damage pitches your view"};
@@ -311,6 +313,32 @@ extern matrix4x4_t viewmodelmatrix;
 #include "cl_collision.h"
 #include "csprogs.h"
 
+/*
+==================
+CL_StairSmoothing
+
+==================
+*/
+void CL_StairSmoothing (void)
+{
+	if (v_dmg_time > 0)
+		v_dmg_time -= (cl.time - cl.oldtime);
+
+	// stair smoothing
+	if (cl.onground && cl.stairoffset < 0)
+	{
+		cl.stairoffset += (cl.time - cl.oldtime) * cl_stairsmoothspeed.value;
+		cl.stairoffset = bound(-16, cl.stairoffset, 0);
+	}
+	else if (cl.onground && cl.stairoffset > 0)
+	{
+		cl.stairoffset -= (cl.time - cl.oldtime) * cl_stairsmoothspeed.value;
+		cl.stairoffset = bound(0, cl.stairoffset, 16);
+	}
+	else
+		cl.stairoffset = 0;
+}
+
 /*
 ==================
 V_CalcRefdef
@@ -336,6 +364,13 @@ void V_CalcRefdef (void)
 		// and the angles from the input system
 		Matrix4x4_OriginFromMatrix(&ent->render.matrix, vieworg);
 		VectorCopy(cl.viewangles, viewangles);
+
+		// update the stairoffset if the player entity has gone up or down without leaving the ground
+		//Con_Printf("cl.onground %i oldz %f newz %f vel %f %f %f\n", cl.onground, oldz, vieworg[2], cl.movement_velocity[0], cl.movement_velocity[1], cl.movement_velocity[2]);
+		cl.stairoffset -= vieworg[2] - oldz;
+		oldz = vieworg[2];
+		cl.stairoffset = bound(-16, cl.stairoffset, 16);
+
 		// interpolate the angles if playing a demo or spectating someone
 		if (cls.demoplayback || cl.fixangle[0])
 		{
@@ -359,7 +394,7 @@ void V_CalcRefdef (void)
 			else
 			{
 				r_view.matrix = ent->render.matrix;
-				r_view.matrix.m[2][3] += cl.stats[STAT_VIEWHEIGHT];
+				Matrix4x4_AdjustOrigin(&r_view.matrix, 0, 0, cl.stats[STAT_VIEWHEIGHT]);
 			}
 			viewmodelmatrix = r_view.matrix;
 		}
@@ -369,28 +404,8 @@ void V_CalcRefdef (void)
 			// TODO: add a cvar to disable this
 			viewangles[PITCH] += cl.qw_weaponkick;
 
-			if (cl.onground)
-			{
-				if (!cl.oldonground)
-					cl.hitgroundtime = cl.time;
-				cl.lastongroundtime = cl.time;
-			}
-			cl.oldonground = cl.onground;
-
-			// stair smoothing
-			//Con_Printf("cl.onground %i oldz %f newz %f\n", cl.onground, oldz, vieworg[2]);
-			if (cl.onground && oldz < vieworg[2])
-			{
-				oldz += (cl.time - cl.oldtime) * cl_stairsmoothspeed.value;
-				oldz = vieworg[2] = bound(vieworg[2] - 16, oldz, vieworg[2]);
-			}
-			else if (cl.onground && oldz > vieworg[2])
-			{
-				oldz -= (cl.time - cl.oldtime) * cl_stairsmoothspeed.value;
-				oldz = vieworg[2] = bound(vieworg[2], oldz, vieworg[2] + 16);
-			}
-			else
-				oldz = vieworg[2];
+			// bias by stair smoothing offset
+			vieworg[2] += cl.stairoffset;
 
 			if (chase_active.value)
 			{
@@ -435,7 +450,6 @@ void V_CalcRefdef (void)
 				{
 					viewangles[ROLL] += v_dmg_time/v_kicktime.value*v_dmg_roll;
 					viewangles[PITCH] += v_dmg_time/v_kicktime.value*v_dmg_pitch;
-					v_dmg_time -= cl.realframetime;
 				}
 				// origin
 				VectorAdd(vieworg, cl.punchvector, vieworg);
@@ -515,7 +529,7 @@ void V_CalcRefdef (void)
 			else
 				Matrix4x4_CreateFromQuakeEntity(&r_view.matrix, vieworg[0], vieworg[1], vieworg[2], viewangles[0], viewangles[1], viewangles[2] + v_idlescale.value * sin(cl.time*v_iroll_cycle.value) * v_iroll_level.value, 1);
 			// calculate a viewmodel matrix for use in view-attached entities
-			Matrix4x4_CreateFromQuakeEntity(&viewmodelmatrix, gunorg[0], gunorg[1], gunorg[2], viewangles[0], viewangles[1], viewangles[2], 0.3);
+			Matrix4x4_CreateFromQuakeEntity(&viewmodelmatrix, gunorg[0], gunorg[1], gunorg[2], viewangles[0], viewangles[1], viewangles[2], cl_viewmodel_scale.value);
 			VectorCopy(vieworg, csqc_origin);
 			VectorCopy(viewangles, csqc_angles);
 		}
@@ -682,6 +696,8 @@ void V_Init (void)
 	Cvar_RegisterVariable (&cl_bobmodel_up);
 	Cvar_RegisterVariable (&cl_bobmodel_speed);
 
+	Cvar_RegisterVariable (&cl_viewmodel_scale);
+
 	Cvar_RegisterVariable (&v_kicktime);
 	Cvar_RegisterVariable (&v_kickroll);
 	Cvar_RegisterVariable (&v_kickpitch);
